# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import error
from twisted.trial import unittest

from buildbot.process import remotetransfer
from buildbot.process.results import FAILURE
from buildbot.process.results import RETRY
from buildbot.process.results import SUCCESS
from buildbot.steps.source import mtn
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectCpdir
from buildbot.test.steps import ExpectDownloadFile
from buildbot.test.steps import ExpectRemoteRef
from buildbot.test.steps import ExpectRmdir
from buildbot.test.steps import ExpectShell
from buildbot.test.steps import ExpectStat
from buildbot.test.util import config
from buildbot.test.util import sourcesteps


class TestMonotone(
    sourcesteps.SourceStepMixin, config.ConfigErrorsMixin, TestReactorMixin, unittest.TestCase
):
    # Just some random revision id to test.
    REVID = '95215e2a9a9f8b6f5c9664e3807cd34617ea928c'
    MTN_VER = 'monotone 1.0 (base revision: UNKNOWN_REV)'

    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_mode_full_clean(self):
        self.setup_step(
            mtn.Monotone(
                repourl='mtn://localhost/monotone', mode='full', method='clean', branch='master'
            )
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'info', '--db', 'db.mtn'])
            .stdout('')
            .exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/_MTN', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'ls', 'unknown'])
            .stdout('file1\nfile2')
            .exit(0),
            ExpectRmdir(dir=['wkdir/file1', 'wkdir/file2'], log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['mtn', 'update', '--revision', 'h:master', '--branch', 'master'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'automate', 'select', 'w:'])
            .stdout(self.REVID)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', self.REVID, 'Monotone')
        return self.run_step()

    def test_mode_full_clean_patch(self):
        self.setup_step(
            mtn.Monotone(
                repourl='mtn://localhost/monotone', mode='full', method='clean', branch='master'
            ),
            patch=(1, 'patch'),
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'info', '--db', 'db.mtn'])
            .stdout('')
            .exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/_MTN', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'ls', 'unknown'])
            .stdout('file1\nfile2')
            .exit(0),
            ExpectRmdir(dir=['wkdir/file1', 'wkdir/file2'], log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['mtn', 'update', '--revision', 'h:master', '--branch', 'master'],
            ).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                workerdest='.buildbot-diff',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                workerdest='.buildbot-patched',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'patch',
                    '-p1',
                    '--remove-empty-files',
                    '--force',
                    '--forward',
                    '-i',
                    '.buildbot-diff',
                ],
            ).exit(0),
            ExpectRmdir(dir='wkdir/.buildbot-diff', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'automate', 'select', 'w:'])
            .stdout(self.REVID)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', self.REVID, 'Monotone')
        return self.run_step()

    def test_mode_full_clean_patch_worker_2_16(self):
        self.setup_build(worker_version={'*': '2.16'})
        self.setup_step(
            mtn.Monotone(
                repourl='mtn://localhost/monotone', mode='full', method='clean', branch='master'
            ),
            patch=(1, 'patch'),
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'info', '--db', 'db.mtn'])
            .stdout('')
            .exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/_MTN', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'ls', 'unknown'])
            .stdout('file1\nfile2')
            .exit(0),
            ExpectRmdir(dir=['wkdir/file1', 'wkdir/file2'], log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['mtn', 'update', '--revision', 'h:master', '--branch', 'master'],
            ).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                slavedest='.buildbot-diff',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                slavedest='.buildbot-patched',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'patch',
                    '-p1',
                    '--remove-empty-files',
                    '--force',
                    '--forward',
                    '-i',
                    '.buildbot-diff',
                ],
            ).exit(0),
            ExpectRmdir(dir='wkdir/.buildbot-diff', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'automate', 'select', 'w:'])
            .stdout(self.REVID)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', self.REVID, 'Monotone')
        return self.run_step()

    def test_mode_full_clean_patch_fail(self):
        self.setup_step(
            mtn.Monotone(
                repourl='mtn://localhost/monotone', mode='full', method='clean', branch='master'
            ),
            patch=(1, 'patch'),
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'info', '--db', 'db.mtn'])
            .stdout('')
            .exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/_MTN', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'ls', 'unknown'])
            .stdout('file1\nfile2')
            .exit(0),
            ExpectRmdir(dir=['wkdir/file1', 'wkdir/file2'], log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['mtn', 'update', '--revision', 'h:master', '--branch', 'master'],
            ).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                workerdest='.buildbot-diff',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                workerdest='.buildbot-patched',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'patch',
                    '-p1',
                    '--remove-empty-files',
                    '--force',
                    '--forward',
                    '-i',
                    '.buildbot-diff',
                ],
            ).exit(0),
            ExpectRmdir(dir='wkdir/.buildbot-diff', log_environ=True).exit(1),
        )
        self.expect_outcome(result=FAILURE, state_string="update (failure)")
        return self.run_step()

    def test_mode_full_clean_no_existing_db(self):
        self.setup_step(
            mtn.Monotone(
                repourl='mtn://localhost/monotone', mode='full', method='clean', branch='master'
            )
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(1),
            ExpectShell(workdir='.', command=['mtn', 'db', 'init', '--db', 'db.mtn']).exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/_MTN', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'ls', 'unknown'])
            .stdout('file1\nfile2')
            .exit(0),
            ExpectRmdir(dir=['wkdir/file1', 'wkdir/file2'], log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['mtn', 'update', '--revision', 'h:master', '--branch', 'master'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'automate', 'select', 'w:'])
            .stdout(self.REVID)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', self.REVID, 'Monotone')
        return self.run_step()

    def test_mode_full_clean_no_existing_checkout(self):
        self.setup_step(
            mtn.Monotone(
                repourl='mtn://localhost/monotone', mode='full', method='clean', branch='master'
            )
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'info', '--db', 'db.mtn'])
            .stdout('')
            .exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/_MTN', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='.',
                command=['mtn', 'checkout', 'wkdir', '--db', 'db.mtn', '--branch', 'master'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'automate', 'select', 'w:'])
            .stdout(self.REVID)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', self.REVID, 'Monotone')
        return self.run_step()

    def test_mode_full_clean_from_scratch(self):
        self.setup_step(
            mtn.Monotone(
                repourl='mtn://localhost/monotone', mode='full', method='clean', branch='master'
            )
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(1),
            ExpectShell(workdir='.', command=['mtn', 'db', 'init', '--db', 'db.mtn']).exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/_MTN', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='.',
                command=['mtn', 'checkout', 'wkdir', '--db', 'db.mtn', '--branch', 'master'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'automate', 'select', 'w:'])
            .stdout(self.REVID)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', self.REVID, 'Monotone')
        return self.run_step()

    def test_mode_full_clobber(self):
        self.setup_step(
            mtn.Monotone(
                repourl='mtn://localhost/monotone', mode='full', method='clobber', branch='master'
            )
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'info', '--db', 'db.mtn'])
            .stdout('')
            .exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='.',
                command=['mtn', 'checkout', 'wkdir', '--db', 'db.mtn', '--branch', 'master'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'automate', 'select', 'w:'])
            .stdout(self.REVID)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', self.REVID, 'Monotone')
        return self.run_step()

    def test_mode_full_clobber_no_existing_db(self):
        self.setup_step(
            mtn.Monotone(
                repourl='mtn://localhost/monotone', mode='full', method='clobber', branch='master'
            )
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(1),
            ExpectShell(workdir='.', command=['mtn', 'db', 'init', '--db', 'db.mtn']).exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='.',
                command=['mtn', 'checkout', 'wkdir', '--db', 'db.mtn', '--branch', 'master'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'automate', 'select', 'w:'])
            .stdout(self.REVID)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', self.REVID, 'Monotone')
        return self.run_step()

    def test_mode_incremental_no_existing_db(self):
        self.setup_step(
            mtn.Monotone(repourl='mtn://localhost/monotone', mode='incremental', branch='master')
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(1),
            ExpectShell(workdir='.', command=['mtn', 'db', 'init', '--db', 'db.mtn']).exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/_MTN', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['mtn', 'update', '--revision', 'h:master', '--branch', 'master'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'automate', 'select', 'w:'])
            .stdout(self.REVID)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', self.REVID, 'Monotone')
        return self.run_step()

    def test_mode_incremental_no_existing_checkout(self):
        self.setup_step(
            mtn.Monotone(repourl='mtn://localhost/monotone', mode='incremental', branch='master')
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'info', '--db', 'db.mtn'])
            .stdout('')
            .exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/_MTN', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='.',
                command=['mtn', 'checkout', 'wkdir', '--db', 'db.mtn', '--branch', 'master'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'automate', 'select', 'w:'])
            .stdout(self.REVID)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', self.REVID, 'Monotone')
        return self.run_step()

    def test_mode_incremental_from_scratch(self):
        self.setup_step(
            mtn.Monotone(repourl='mtn://localhost/monotone', mode='incremental', branch='master')
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(1),
            ExpectShell(workdir='.', command=['mtn', 'db', 'init', '--db', 'db.mtn']).exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/_MTN', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='.',
                command=['mtn', 'checkout', 'wkdir', '--db', 'db.mtn', '--branch', 'master'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'automate', 'select', 'w:'])
            .stdout(self.REVID)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', self.REVID, 'Monotone')
        return self.run_step()

    def test_mode_incremental(self):
        self.setup_step(
            mtn.Monotone(repourl='mtn://localhost/monotone', mode='incremental', branch='master')
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'info', '--db', 'db.mtn'])
            .stdout('')
            .exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/_MTN', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['mtn', 'update', '--revision', 'h:master', '--branch', 'master'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'automate', 'select', 'w:'])
            .stdout(self.REVID)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', self.REVID, 'Monotone')
        return self.run_step()

    def test_mode_incremental_retry(self):
        self.setup_step(
            mtn.Monotone(
                repourl='mtn://localhost/monotone',
                mode='incremental',
                branch='master',
                retry=(0, 1),
            )
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'info', '--db', 'db.mtn'])
            .stdout('')
            .exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(1),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/_MTN', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['mtn', 'update', '--revision', 'h:master', '--branch', 'master'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'automate', 'select', 'w:'])
            .stdout(self.REVID)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', self.REVID, 'Monotone')
        return self.run_step()

    def test_mode_full_fresh(self):
        self.setup_step(
            mtn.Monotone(
                repourl='mtn://localhost/monotone', mode='full', method='fresh', branch='master'
            )
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'info', '--db', 'db.mtn'])
            .stdout('')
            .exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/_MTN', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'ls', 'unknown'])
            .stdout('file1\nfile2')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'ls', 'ignored'])
            .stdout('file3\nfile4')
            .exit(0),
            ExpectRmdir(
                dir=['wkdir/file1', 'wkdir/file2', 'wkdir/file3', 'wkdir/file4'], log_environ=True
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['mtn', 'update', '--revision', 'h:master', '--branch', 'master'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'automate', 'select', 'w:'])
            .stdout(self.REVID)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', self.REVID, 'Monotone')
        return self.run_step()

    def test_mode_incremental_given_revision(self):
        self.setup_step(
            mtn.Monotone(repourl='mtn://localhost/monotone', mode='incremental', branch='master'),
            {"revision": 'abcdef01'},
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'info', '--db', 'db.mtn'])
            .stdout('')
            .exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/_MTN', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['mtn', 'update', '--revision', 'abcdef01', '--branch', 'master'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'automate', 'select', 'w:'])
            .stdout('abcdef019a9f8b6f5c9664e3807cd34617ea928c')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', 'abcdef019a9f8b6f5c9664e3807cd34617ea928c', 'Monotone')
        return self.run_step()

    def test_mode_full_copy(self):
        self.setup_step(
            mtn.Monotone(
                repourl='mtn://localhost/monotone', mode='full', method='copy', branch='master'
            )
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'info', '--db', 'db.mtn'])
            .stdout('')
            .exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectStat(file='source/_MTN', log_environ=True).exit(0),
            ExpectShell(
                workdir='source',
                command=['mtn', 'update', '--revision', 'h:master', '--branch', 'master'],
            ).exit(0),
            ExpectCpdir(fromdir='source', todir='build', log_environ=True, timeout=1200).exit(0),
            ExpectShell(workdir='build', command=['mtn', 'automate', 'select', 'w:'])
            .stdout(self.REVID)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', self.REVID, 'Monotone')
        return self.run_step()

    def test_mode_full_no_method(self):
        self.setup_step(
            mtn.Monotone(repourl='mtn://localhost/monotone', mode='full', branch='master')
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'info', '--db', 'db.mtn'])
            .stdout('')
            .exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectStat(file='source/_MTN', log_environ=True).exit(0),
            ExpectShell(
                workdir='source',
                command=['mtn', 'update', '--revision', 'h:master', '--branch', 'master'],
            ).exit(0),
            ExpectCpdir(fromdir='source', todir='build', log_environ=True, timeout=1200).exit(0),
            ExpectShell(workdir='build', command=['mtn', 'automate', 'select', 'w:'])
            .stdout(self.REVID)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', self.REVID, 'Monotone')
        return self.run_step()

    def test_incorrect_method(self):
        with self.assertRaisesConfigError("Invalid method for mode == full"):
            mtn.Monotone(
                repourl='mtn://localhost/monotone',
                mode='full',
                method='wrongmethod',
                branch='master',
            )

    def test_incremental_invalid_method(self):
        with self.assertRaisesConfigError("Incremental mode does not require method"):
            mtn.Monotone(
                repourl='mtn://localhost/monotone',
                mode='incremental',
                method='fresh',
                branch="master",
            )

    def test_repourl(self):
        with self.assertRaisesConfigError("must provide repourl"):
            mtn.Monotone(mode="full", branch="master")

    def test_branch(self):
        with self.assertRaisesConfigError("must provide branch"):
            mtn.Monotone(
                repourl='mtn://localhost/monotone',
                mode="full",
            )

    def test_mode_incremental_patched(self):
        self.setup_step(
            mtn.Monotone(repourl='mtn://localhost/monotone', mode='incremental', branch='master')
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'info', '--db', 'db.mtn'])
            .stdout('')
            .exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'ls', 'unknown'])
            .stdout('file1\nfile2')
            .exit(0),
            ExpectRmdir(dir=['wkdir/file1', 'wkdir/file2'], log_environ=True).exit(0),
            ExpectStat(file='wkdir/_MTN', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['mtn', 'update', '--revision', 'h:master', '--branch', 'master'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'automate', 'select', 'w:'])
            .stdout(self.REVID)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', self.REVID, 'Monotone')
        return self.run_step()

    def test_worker_connection_lost(self):
        self.setup_step(
            mtn.Monotone(
                repourl='mtn://localhost/monotone', mode='full', method='clean', branch='master'
            )
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version'])
            .stdout(self.MTN_VER)
            .error(error.ConnectionLost())
        )
        self.expect_outcome(result=RETRY, state_string="update (retry)")
        return self.run_step()

    def test_database_migration(self):
        self.setup_step(
            mtn.Monotone(repourl='mtn://localhost/monotone', mode='incremental', branch='master')
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'info', '--db', 'db.mtn'])
            .stdout('migration needed')
            .exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'migrate', '--db', 'db.mtn']).exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/_MTN', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['mtn', 'update', '--revision', 'h:master', '--branch', 'master'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'automate', 'select', 'w:'])
            .stdout(self.REVID)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', self.REVID, 'Monotone')
        return self.run_step()

    def test_database_invalid(self):
        self.setup_step(
            mtn.Monotone(repourl='mtn://localhost/monotone', mode='incremental', branch='master')
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'info', '--db', 'db.mtn'])
            .stdout('not a monotone database')
            .exit(0),
        )
        self.expect_outcome(result=FAILURE)
        return self.run_step()

    def test_database_too_new(self):
        self.setup_step(
            mtn.Monotone(repourl='mtn://localhost/monotone', mode='incremental', branch='master')
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'info', '--db', 'db.mtn'])
            .stdout('too new, cannot use')
            .exit(0),
            ExpectRmdir(dir='db.mtn', log_environ=True).exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'init', '--db', 'db.mtn']).exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/_MTN', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['mtn', 'update', '--revision', 'h:master', '--branch', 'master'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'automate', 'select', 'w:'])
            .stdout(self.REVID)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', self.REVID, 'Monotone')
        return self.run_step()

    def test_database_empty(self):
        self.setup_step(
            mtn.Monotone(repourl='mtn://localhost/monotone', mode='incremental', branch='master')
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['mtn', '--version']).stdout(self.MTN_VER).exit(0),
            ExpectStat(file='db.mtn', log_environ=True).exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'info', '--db', 'db.mtn'])
            .stdout('database has no tables')
            .exit(0),
            ExpectRmdir(dir='db.mtn', log_environ=True).exit(0),
            ExpectShell(workdir='.', command=['mtn', 'db', 'init', '--db', 'db.mtn']).exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'mtn',
                    'pull',
                    'mtn://localhost/monotone?master',
                    '--db',
                    'db.mtn',
                    '--ticker=dot',
                ],
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/_MTN', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['mtn', 'update', '--revision', 'h:master', '--branch', 'master'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['mtn', 'automate', 'select', 'w:'])
            .stdout(self.REVID)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', self.REVID, 'Monotone')
        return self.run_step()



#!/usr/bin/env python
#
# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

"""
Standard setup script.
"""

import os
import sys

from setuptools import Command
from setuptools import setup
from setuptools.command.sdist import sdist

from buildbot_worker import version

BUILDING_WHEEL = bool("bdist_wheel" in sys.argv)


class our_install_data(Command):
    def initialize_options(self):
        self.install_dir = None

    def finalize_options(self):
        self.set_undefined_options(
            'install',
            ('install_lib', 'install_dir'),
        )

    def run(self):
        # ensure there's a buildbot_worker/VERSION file
        fn = os.path.join(self.install_dir, 'buildbot_worker', 'VERSION')
        with open(fn, 'w') as f:
            f.write(version)


class our_sdist(sdist):
    def make_release_tree(self, base_dir, files):
        sdist.make_release_tree(self, base_dir, files)
        # ensure there's a buildbot_worker/VERSION file
        fn = os.path.join(base_dir, 'buildbot_worker', 'VERSION')
        open(fn, 'w').write(version)

        # ensure that NEWS has a copy of the latest release notes, copied from
        # the master tree, with the proper version substituted
        src_fn = os.path.join('..', 'master', 'docs', 'relnotes/index.rst')
        with open(src_fn) as f:
            src = f.read()
        src = src.replace('|version|', version)
        dst_fn = os.path.join(base_dir, 'NEWS')
        with open(dst_fn, 'w') as f:
            f.write(src)


setup_args = {
    'name': "buildbot-worker",
    'version': version,
    'description': "Buildbot Worker Daemon",
    'long_description': "See the 'buildbot' package for details",
    'author': "Brian Warner",
    'author_email': "warner-buildbot@lothar.com",
    'maintainer': "Dustin J. Mitchell",
    'maintainer_email': "dustin@v.igoro.us",
    'url': "http://buildbot.net/",
    'classifiers': [
        'Development Status :: 5 - Production/Stable',
        'Environment :: No Input/Output (Daemon)',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Testing',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ],
    'packages': [
        "buildbot_worker",
        "buildbot_worker.util",
        "buildbot_worker.commands",
        "buildbot_worker.scripts",
        "buildbot_worker.monkeypatches",
    ]
    + (
        []
        if BUILDING_WHEEL
        else [  # skip tests for wheels (save 40% of the archive)
            "buildbot_worker.test",
            "buildbot_worker.test.fake",
            "buildbot_worker.test.unit",
            "buildbot_worker.test.util",
        ]
    ),
    # mention data_files, even if empty, so install_data is called and
    # VERSION gets copied
    'data_files': [("buildbot_worker", [])],
    'package_data': {
        '': [
            'VERSION',
        ]
    },
    'cmdclass': {'install_data': our_install_data, 'sdist': our_sdist},
    'entry_points': {
        'console_scripts': [
            'buildbot-worker=buildbot_worker.scripts.runner:run',
            # this will also be shipped on non windows :-(
            'buildbot_worker_windows_service=buildbot_worker.scripts.windows_service:HandleCommandLine',
        ]
    },
}

# set zip_safe to false to force Windows installs to always unpack eggs
# into directories, which seems to work better --
# see http://buildbot.net/trac/ticket/907
if sys.platform == "win32":
    setup_args['zip_safe'] = False

twisted_ver = ">= 21.2.0"

setup_args['install_requires'] = [
    'twisted ' + twisted_ver,
]

setup_args['install_requires'] += [
    'autobahn >= 0.16.0',
    'msgpack >= 0.6.0',
]

# buildbot_worker_windows_service needs pywin32
if sys.platform == "win32":
    setup_args['install_requires'].append('pywin32')

# Unit test hard dependencies.
test_deps = [
    'psutil',
]

setup_args['tests_require'] = test_deps

setup_args['extras_require'] = {
    'test': test_deps,
}

if '--help-commands' in sys.argv or 'trial' in sys.argv or 'test' in sys.argv:
    setup_args['setup_requires'] = [
        'setuptools_trial',
    ]

if os.getenv('NO_INSTALL_REQS'):
    setup_args['install_requires'] = None
    setup_args['extras_require'] = None

setup(**setup_args)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import datetime

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.process.properties import Interpolate
from buildbot.process.properties import Properties
from buildbot.process.properties import renderer
from buildbot.process.results import FAILURE
from buildbot.process.results import SUCCESS
from buildbot.reporters.generators.build import BuildStartEndStatusGenerator
from buildbot.reporters.gerrit_verify_status import GerritVerifyStatusPush
from buildbot.reporters.message import MessageFormatterRenderable
from buildbot.test.fake import fakemaster
from buildbot.test.fake import httpclientservice as fakehttpclientservice
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import logging
from buildbot.test.util.config import ConfigErrorsMixin
from buildbot.test.util.reporter import ReporterTestMixin


class TestGerritVerifyStatusPush(
    TestReactorMixin, ReporterTestMixin, ConfigErrorsMixin, logging.LoggingMixin, unittest.TestCase
):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.setup_reporter_test()
        self.reporter_test_props = {'gerrit_changes': [{'change_id': 12, 'revision_id': 2}]}

        self.master = yield fakemaster.make_master(self, wantData=True, wantDb=True, wantMq=True)

        yield self.master.startService()
        self.addCleanup(self.master.stopService)

    @defer.inlineCallbacks
    def createGerritStatus(self, **kwargs):
        auth = kwargs.pop('auth', ('log', Interpolate('pass')))

        self._http = yield fakehttpclientservice.HTTPClientService.getService(
            self.master, self, "gerrit", auth=('log', 'pass'), debug=None, verify=None
        )
        self.sp = GerritVerifyStatusPush("gerrit", auth=auth, **kwargs)
        yield self.sp.setServiceParent(self.master)

    @defer.inlineCallbacks
    def test_basic(self):
        yield self.createGerritStatus()
        build = yield self.insert_build_new()
        self._http.expect(
            method='post',
            ep='/a/changes/12/revisions/2/verify-status~verifications',
            json={
                'comment': 'Build started.',
                'abstain': False,
                'name': 'Builder0',
                'reporter': 'buildbot',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'value': 0,
                'duration': 'pending',
            },
        )
        self._http.expect(
            method='post',
            ep='/a/changes/12/revisions/2/verify-status~verifications',
            json={
                'comment': 'Build done.',
                'abstain': False,
                'name': 'Builder0',
                'reporter': 'buildbot',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'value': 1,
                'duration': '2h 1m 4s',
            },
        )
        self._http.expect(
            method='post',
            ep='/a/changes/12/revisions/2/verify-status~verifications',
            json={
                'comment': 'Build done.',
                'abstain': False,
                'name': 'Builder0',
                'reporter': 'buildbot',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'value': -1,
                'duration': '2h 1m 4s',
            },
        )
        yield self.sp._got_event(('builds', 20, 'new'), build)
        build['complete'] = True
        build['complete_at'] = build['started_at'] + datetime.timedelta(
            hours=2, minutes=1, seconds=4
        )
        build['results'] = SUCCESS
        yield self.sp._got_event(('builds', 20, 'finished'), build)
        build['results'] = FAILURE
        yield self.sp._got_event(('builds', 20, 'finished'), build)

    @defer.inlineCallbacks
    def test_custom_description(self):
        start_formatter = MessageFormatterRenderable(Interpolate("started %(prop:buildername)s"))
        end_formatter = MessageFormatterRenderable(Interpolate("finished %(prop:buildername)s"))

        generator = BuildStartEndStatusGenerator(
            start_formatter=start_formatter, end_formatter=end_formatter
        )

        yield self.createGerritStatus(generators=[generator])
        build = yield self.insert_build_new()
        self._http.expect(
            method='post',
            ep='/a/changes/12/revisions/2/verify-status~verifications',
            json={
                'comment': 'started Builder0',
                'abstain': False,
                'name': 'Builder0',
                'reporter': 'buildbot',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'value': 0,
                'duration': 'pending',
            },
        )
        self._http.expect(
            method='post',
            ep='/a/changes/12/revisions/2/verify-status~verifications',
            json={
                'comment': 'finished Builder0',
                'abstain': False,
                'name': 'Builder0',
                'reporter': 'buildbot',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'value': 1,
                'duration': '2h 1m 4s',
            },
        )
        yield self.sp._got_event(('builds', 20, 'new'), build)
        build['complete'] = True
        build['complete_at'] = build['started_at'] + datetime.timedelta(
            hours=2, minutes=1, seconds=4
        )
        build['results'] = SUCCESS
        yield self.sp._got_event(('builds', 20, 'finished'), build)

    @defer.inlineCallbacks
    def test_custom_name(self):
        yield self.createGerritStatus(verification_name=Interpolate("builder %(prop:buildername)s"))
        build = yield self.insert_build_new()
        self._http.expect(
            method='post',
            ep='/a/changes/12/revisions/2/verify-status~verifications',
            json={
                'comment': 'Build started.',
                'abstain': False,
                'name': 'builder Builder0',
                'reporter': 'buildbot',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'value': 0,
                'duration': 'pending',
            },
        )
        self._http.expect(
            method='post',
            ep='/a/changes/12/revisions/2/verify-status~verifications',
            json={
                'comment': 'Build done.',
                'abstain': False,
                'name': 'builder Builder0',
                'reporter': 'buildbot',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'value': 1,
                'duration': '2h 1m 4s',
            },
        )
        yield self.sp._got_event(('builds', 20, 'new'), build)
        build['complete'] = True
        build['complete_at'] = build['started_at'] + datetime.timedelta(
            hours=2, minutes=1, seconds=4
        )
        build['results'] = SUCCESS
        yield self.sp._got_event(('builds', 20, 'finished'), build)

    @defer.inlineCallbacks
    def test_custom_abstain(self):
        yield self.createGerritStatus(
            abstain=renderer(lambda p: p.getProperty("buildername") == 'Builder0')
        )
        build = yield self.insert_build_new()
        self._http.expect(
            method='post',
            ep='/a/changes/12/revisions/2/verify-status~verifications',
            json={
                'comment': 'Build started.',
                'abstain': True,
                'name': 'Builder0',
                'reporter': 'buildbot',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'value': 0,
                'duration': 'pending',
            },
        )
        self._http.expect(
            method='post',
            ep='/a/changes/12/revisions/2/verify-status~verifications',
            json={
                'comment': 'Build done.',
                'abstain': True,
                'name': 'Builder0',
                'reporter': 'buildbot',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'value': 1,
                'duration': '2h 1m 4s',
            },
        )
        yield self.sp._got_event(('builds', 20, 'new'), build)
        build['complete'] = True
        build['complete_at'] = build['started_at'] + datetime.timedelta(
            hours=2, minutes=1, seconds=4
        )
        build['results'] = SUCCESS
        yield self.sp._got_event(('builds', 20, 'finished'), build)

    @defer.inlineCallbacks
    def test_custom_category(self):
        yield self.createGerritStatus(category=renderer(lambda p: p.getProperty("buildername")))
        build = yield self.insert_build_new()
        self._http.expect(
            method='post',
            ep='/a/changes/12/revisions/2/verify-status~verifications',
            json={
                'comment': 'Build started.',
                'abstain': False,
                'category': 'Builder0',
                'name': 'Builder0',
                'reporter': 'buildbot',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'value': 0,
                'duration': 'pending',
            },
        )
        self._http.expect(
            method='post',
            ep='/a/changes/12/revisions/2/verify-status~verifications',
            json={
                'comment': 'Build done.',
                'abstain': False,
                'category': 'Builder0',
                'name': 'Builder0',
                'reporter': 'buildbot',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'value': 1,
                'duration': '2h 1m 4s',
            },
        )
        yield self.sp._got_event(('builds', 20, 'new'), build)
        build['complete'] = True
        build['complete_at'] = build['started_at'] + datetime.timedelta(
            hours=2, minutes=1, seconds=4
        )
        build['results'] = SUCCESS
        yield self.sp._got_event(('builds', 20, 'finished'), build)

    @defer.inlineCallbacks
    def test_custom_reporter(self):
        yield self.createGerritStatus(reporter=renderer(lambda p: p.getProperty("buildername")))
        build = yield self.insert_build_new()
        self._http.expect(
            method='post',
            ep='/a/changes/12/revisions/2/verify-status~verifications',
            json={
                'comment': 'Build started.',
                'abstain': False,
                'name': 'Builder0',
                'reporter': 'Builder0',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'value': 0,
                'duration': 'pending',
            },
        )
        self._http.expect(
            method='post',
            ep='/a/changes/12/revisions/2/verify-status~verifications',
            json={
                'comment': 'Build done.',
                'abstain': False,
                'name': 'Builder0',
                'reporter': 'Builder0',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'value': 1,
                'duration': '2h 1m 4s',
            },
        )
        yield self.sp._got_event(('builds', 20, 'new'), build)
        build['complete'] = True
        build['complete_at'] = build['started_at'] + datetime.timedelta(
            hours=2, minutes=1, seconds=4
        )
        build['results'] = SUCCESS
        yield self.sp._got_event(('builds', 20, 'finished'), build)

    @defer.inlineCallbacks
    def test_verbose(self):
        yield self.createGerritStatus(verbose=True)
        build = yield self.insert_build_new()
        self._http.expect(
            method='post',
            ep='/a/changes/12/revisions/2/verify-status~verifications',
            json={
                'comment': 'Build started.',
                'abstain': False,
                'name': 'Builder0',
                'reporter': 'buildbot',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'value': 0,
                'duration': 'pending',
            },
        )
        self.setUpLogging()
        yield self.sp._got_event(('builds', 20, 'new'), build)
        self.assertLogged("Sending Gerrit status for")

    @defer.inlineCallbacks
    def test_not_verbose(self):
        yield self.createGerritStatus(verbose=False)
        build = yield self.insert_build_new()
        self._http.expect(
            method='post',
            ep='/a/changes/12/revisions/2/verify-status~verifications',
            json={
                'comment': 'Build started.',
                'abstain': False,
                'name': 'Builder0',
                'reporter': 'buildbot',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'value': 0,
                'duration': 'pending',
            },
        )
        self.setUpLogging()
        self._http.quiet = True
        yield self.sp._got_event(('builds', 20, 'new'), build)
        self.assertWasQuiet()

    @defer.inlineCallbacks
    def test_format_duration(self):
        yield self.createGerritStatus(verbose=False)
        self.assertEqual(self.sp.formatDuration(datetime.timedelta(seconds=1)), "0m 1s")
        self.assertEqual(self.sp.formatDuration(datetime.timedelta(hours=1, seconds=1)), "1h 0m 1s")
        self.assertEqual(
            self.sp.formatDuration(datetime.timedelta(days=1, seconds=1)), "1 day 0h 0m 1s"
        )
        self.assertEqual(
            self.sp.formatDuration(datetime.timedelta(days=2, seconds=1)), "2 days 0h 0m 1s"
        )

    @defer.inlineCallbacks
    def test_gerrit_changes(self):
        yield self.createGerritStatus()

        # from chdict:
        change_props = {
            'event.change.owner.email': 'dustin@mozilla.com',
            'event.change.subject': 'fix 1234',
            'event.change.project': 'pr',
            'event.change.owner.name': 'Dustin',
            'event.change.number': '4321',
            'event.change.url': 'http://buildbot.net',
            'event.change.branch': 'br',
            'event.type': 'patchset-created',
            'event.patchSet.revision': 'abcdef',
            'event.patchSet.number': '12',
            'event.source': 'GerritChangeSource',
        }

        props = Properties.fromDict({k: (v, 'change') for k, v in change_props.items()})
        changes = self.sp.getGerritChanges(props)
        self.assertEqual(changes, [{'change_id': '4321', 'revision_id': '12'}])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import os
import stat
import tempfile
from unittest.mock import Mock

from twisted.trial import unittest

from buildbot.process import remotetransfer


# Test buildbot.steps.remotetransfer.FileWriter class.
class TestFileWriter(unittest.TestCase):
    # test FileWriter.__init__() method.

    def testInit(self):
        #
        # patch functions called in constructor
        #

        # patch os.path.exists() to always return False
        mockedExists = Mock(return_value=False)
        self.patch(os.path, "exists", mockedExists)

        # capture calls to os.makedirs()
        mockedMakedirs = Mock()
        self.patch(os, 'makedirs', mockedMakedirs)

        # capture calls to tempfile.mkstemp()
        mockedMkstemp = Mock(return_value=(7, "tmpname"))
        self.patch(tempfile, "mkstemp", mockedMkstemp)

        # capture calls to os.fdopen()
        mockedFdopen = Mock()
        self.patch(os, "fdopen", mockedFdopen)

        #
        # call _FileWriter constructor
        #
        destfile = os.path.join("dir", "file")
        remotetransfer.FileWriter(destfile, 64, stat.S_IRUSR)

        #
        # validate captured calls
        #
        absdir = os.path.dirname(os.path.abspath(os.path.join("dir", "file")))
        mockedExists.assert_called_once_with(absdir)
        mockedMakedirs.assert_called_once_with(absdir)
        mockedMkstemp.assert_called_once_with(dir=absdir, prefix='buildbot-transfer-')
        mockedFdopen.assert_called_once_with(7, 'wb')


class TestStringFileWriter(unittest.TestCase):
    def testBasic(self):
        sfw = remotetransfer.StringFileWriter()
        # StringFileWriter takes bytes or native string and outputs native strings
        sfw.remote_write(b'bytes')
        sfw.remote_write(' or str')
        self.assertEqual(sfw.buffer, 'bytes or str')



#! /usr/bin/python -tt


from getpass import getpass
from smtplib import SMTP

"""
This script helps to check that the SMTP_HOST (see below) would accept STARTTLS
command, and if LOCAL_HOST is acceptable for it, would check the requested user
name and password would allow to send e-mail through it.
"""


SMTP_HOST = 'the host you want to send e-mail through'
LOCAL_HOST = 'hostname that the SMTP_HOST would accept'


def main():
    """
    entry point
    """

    server = SMTP(SMTP_HOST)

    server.starttls()

    print(server.ehlo(LOCAL_HOST))

    user = input('user: ')
    password = getpass('password: ')

    print(server.login(user, password))
    server.close()


if __name__ == '__main__':
    main()

# vim:ts=4:sw=4:et:tw=80



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

from unittest import mock

from parameterized import parameterized
from twisted.internet import defer
from twisted.trial import unittest

from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.util import service
from buildbot.wamp import connector


class FakeConfig:
    def __init__(self, mq_dict):
        self.mq = mq_dict


class FakeService(service.AsyncMultiService):
    name: str | None = "fakeWampService"  # type: ignore[assignment]
    # Fake wamp service
    # just call the maker on demand by the test

    def __init__(
        self, url, realm, make, extra=None, debug=False, debug_wamp=False, debug_app=False
    ):
        super().__init__()
        self.make = make
        self.extra = extra

    def gotConnection(self):
        self.make(None)
        r = self.make(self)
        r.publish = mock.Mock(spec=r.publish)
        r.register = mock.Mock(spec=r.register)
        r.subscribe = mock.Mock(spec=r.subscribe)
        r.onJoin(None)


class TestedWampConnector(connector.WampConnector):
    serviceClass = FakeService


class WampConnector(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        master = yield fakemaster.make_master(self)
        self.connector = TestedWampConnector()

        config = FakeConfig({'type': 'wamp', 'router_url': "wss://foo", 'realm': "bb"})

        yield self.connector.setServiceParent(master)
        yield master.startService()
        yield self.connector.reconfigServiceWithBuildbotConfig(config)

    @defer.inlineCallbacks
    def test_reconfig_same_config(self):
        config = FakeConfig({'type': 'wamp', 'router_url': "wss://foo", 'realm': "bb"})
        yield self.connector.reconfigServiceWithBuildbotConfig(config)

    @parameterized.expand([
        ('type', 'simple'),
        ('router_url', 'wss://other-foo'),
        ('realm', 'bb-other'),
        ('wamp_debug_level', 'info'),
    ])
    @defer.inlineCallbacks
    def test_reconfig_does_not_allow_config_change(self, attr_name, attr_value):
        mq_dict = {'type': 'wamp', 'router_url': "wss://foo", 'realm': "bb"}
        mq_dict[attr_name] = attr_value
        with self.assertRaises(
            ValueError, msg="Cannot use different wamp settings when reconfiguring"
        ):
            yield self.connector.reconfigServiceWithBuildbotConfig(FakeConfig(mq_dict))

    @defer.inlineCallbacks
    def test_startup(self):
        d = self.connector.getService()
        self.connector.app.gotConnection()
        yield d
        # 824 is the hardcoded masterid of fakemaster
        self.connector.service.publish.assert_called_with("org.buildbot.824.connected")

    @defer.inlineCallbacks
    def test_subscribe(self):
        d = self.connector.subscribe('callback', 'topic', 'options')
        self.connector.app.gotConnection()
        yield d
        self.connector.service.subscribe.assert_called_with('callback', 'topic', 'options')

    @defer.inlineCallbacks
    def test_publish(self):
        d = self.connector.publish('topic', 'data', 'options')
        self.connector.app.gotConnection()
        yield d
        self.connector.service.publish.assert_called_with('topic', 'data', options='options')

    @defer.inlineCallbacks
    def test_OnLeave(self):
        d = self.connector.getService()
        self.connector.app.gotConnection()
        yield d
        self.assertTrue(self.connector.master.running)
        self.connector.service.onLeave(None)
        self.assertFalse(self.connector.master.running)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING

from twisted.internet import defer

from buildbot.data import base
from buildbot.data import masters
from buildbot.data import types
from buildbot.db.schedulers import SchedulerAlreadyClaimedError

if TYPE_CHECKING:
    from buildbot.db.schedulers import SchedulerModel
    from buildbot.util.twisted import InlineCallbacksType


@defer.inlineCallbacks
def _db2data(master, dbdict: SchedulerModel):
    dbmaster = None
    if dbdict.masterid is not None:
        dbmaster = yield master.data.get(('masters', dbdict.masterid))
    data = {
        'schedulerid': dbdict.id,
        'name': dbdict.name,
        'enabled': dbdict.enabled,
        'master': dbmaster,
    }
    return data


class SchedulerEndpoint(base.Endpoint):
    kind = base.EndpointKind.SINGLE
    pathPatterns = [
        "/schedulers/n:schedulerid",
        "/masters/n:masterid/schedulers/n:schedulerid",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        dbdict = yield self.master.db.schedulers.getScheduler(kwargs['schedulerid'])
        if 'masterid' in kwargs:
            if dbdict.masterid != kwargs['masterid']:
                return None
        return (yield _db2data(self.master, dbdict)) if dbdict else None

    @defer.inlineCallbacks
    def control(self, action, args, kwargs):
        if action == 'enable':
            schedulerid = kwargs['schedulerid']
            v = args['enabled']
            yield self.master.data.updates.schedulerEnable(schedulerid, v)
        return None


class SchedulersEndpoint(base.Endpoint):
    kind = base.EndpointKind.COLLECTION
    pathPatterns = [
        "/schedulers",
        "/masters/n:masterid/schedulers",
    ]
    rootLinkName = 'schedulers'

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        schedulers = yield self.master.db.schedulers.getSchedulers(masterid=kwargs.get('masterid'))
        schdicts = yield defer.DeferredList(
            [_db2data(self.master, schdict) for schdict in schedulers],
            consumeErrors=True,
            fireOnOneErrback=True,
        )
        return [r for (s, r) in schdicts]


class Scheduler(base.ResourceType):
    name = "scheduler"
    plural = "schedulers"
    endpoints = [SchedulerEndpoint, SchedulersEndpoint]
    eventPathPatterns = [
        "/schedulers/:schedulerid",
    ]

    class EntityType(types.Entity):
        schedulerid = types.Integer()
        name = types.String()
        enabled = types.Boolean()
        master = types.NoneOk(masters.Master.entityType)

    entityType = EntityType(name)

    @defer.inlineCallbacks
    def generateEvent(self, schedulerid, event):
        scheduler = yield self.master.data.get(('schedulers', str(schedulerid)))
        self.produceEvent(scheduler, event)

    @base.updateMethod
    @defer.inlineCallbacks
    def schedulerEnable(self, schedulerid: int, v: bool) -> InlineCallbacksType[None]:
        yield self.master.db.schedulers.enable(schedulerid, v)
        yield self.generateEvent(schedulerid, 'updated')
        return None

    @base.updateMethod
    def findSchedulerId(self, name: str) -> defer.Deferred[int]:
        return self.master.db.schedulers.findSchedulerId(name)

    @base.updateMethod
    @defer.inlineCallbacks
    def trySetSchedulerMaster(self, schedulerid: int, masterid: int) -> InlineCallbacksType[bool]:
        try:
            yield self.master.db.schedulers.setSchedulerMaster(schedulerid, masterid)
        except SchedulerAlreadyClaimedError:
            return False
        return True

    @defer.inlineCallbacks
    def _masterDeactivated(self, masterid):
        schedulers = yield self.master.db.schedulers.getSchedulers(masterid=masterid)
        for sch in schedulers:
            yield self.master.db.schedulers.setSchedulerMaster(sch.id, None)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.schedulers import timed
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import scheduler


class Timed(scheduler.SchedulerMixin, TestReactorMixin, unittest.TestCase):
    OBJECTID = 928754

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        yield self.setUpScheduler()

    class Subclass(timed.Timed):
        def getNextBuildTime(self, lastActuation):
            self.got_lastActuation = lastActuation
            return defer.succeed((lastActuation or 1000) + 60)

        def startBuild(self):
            self.started_build = True
            return defer.succeed(None)

    @defer.inlineCallbacks
    def makeScheduler(self, firstBuildDuration=0, **kwargs):
        sched = yield self.attachScheduler(self.Subclass(**kwargs), self.OBJECTID)
        return sched

    # tests

    # note that most of the heavy-lifting for testing this class is handled by
    # the subclasses' tests, as that's the more natural place for it



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import re

from parameterized import parameterized
from twisted.trial import unittest

from buildbot.util.ssfilter import SourceStampFilter
from buildbot.util.ssfilter import extract_filter_values
from buildbot.util.ssfilter import extract_filter_values_branch
from buildbot.util.ssfilter import extract_filter_values_regex


class TestSourceStampFilter(unittest.TestCase):
    def test_extract_filter_values(self):
        self.assertEqual(extract_filter_values([], 'name'), [])
        self.assertEqual(extract_filter_values(['value'], 'name'), ['value'])
        self.assertEqual(extract_filter_values('value', 'name'), ['value'])

        with self.assertRaises(ValueError):
            extract_filter_values({'value'}, 'name')
        with self.assertRaises(ValueError):
            extract_filter_values(None, 'name')
        with self.assertRaises(ValueError):
            extract_filter_values([{'value'}], 'name')
        with self.assertRaises(ValueError):
            extract_filter_values([None], 'name')

    def test_extract_filter_values_branch(self):
        self.assertEqual(extract_filter_values_branch([], 'name'), [])
        self.assertEqual(extract_filter_values_branch(['value'], 'name'), ['value'])
        self.assertEqual(extract_filter_values_branch('value', 'name'), ['value'])
        self.assertEqual(extract_filter_values_branch([None], 'name'), [None])
        self.assertEqual(extract_filter_values_branch(None, 'name'), [None])

        with self.assertRaises(ValueError):
            extract_filter_values({'value'}, 'name')
        with self.assertRaises(ValueError):
            extract_filter_values([{'value'}], 'name')

    def test_extract_filter_values_regex(self):
        self.assertEqual(extract_filter_values_regex([], 'name'), [])
        self.assertEqual(extract_filter_values_regex(['value'], 'name'), ['value'])
        self.assertEqual(extract_filter_values_regex('value', 'name'), ['value'])
        self.assertEqual(
            extract_filter_values_regex([re.compile('test')], 'name'), [re.compile('test')]
        )
        self.assertEqual(
            extract_filter_values_regex(re.compile('test'), 'name'), [re.compile('test')]
        )

        with self.assertRaises(ValueError):
            extract_filter_values({'value'}, 'name')
        with self.assertRaises(ValueError):
            extract_filter_values([{'value'}], 'name')

    @parameterized.expand([
        ('match', {'project': 'p', 'codebase': 'c', 'repository': 'r', 'branch': 'b'}, True),
        ('not_project', {'project': '0', 'codebase': 'c', 'repository': 'r', 'branch': 'b'}, False),
        (
            'not_codebase',
            {'project': 'p', 'codebase': '0', 'repository': 'r', 'branch': 'b'},
            False,
        ),
        (
            'not_repository',
            {'project': 'p', 'codebase': 'c', 'repository': '0', 'branch': 'b'},
            False,
        ),
        ('not_branch', {'project': 'p', 'codebase': 'c', 'repository': 'r', 'branch': '0'}, False),
        (
            'none_branch',
            {'project': 'p', 'codebase': 'c', 'repository': 'r', 'branch': None},
            False,
        ),
    ])
    def test_filter_is_matched_eq_or_re(self, name, ss, expected):
        filter = SourceStampFilter(
            project_eq='p', codebase_eq='c', repository_eq='r', branch_eq='b'
        )
        self.assertEqual(filter.is_matched(ss), expected)

        filter = SourceStampFilter(
            project_re='^p$', codebase_re='^c$', repository_re='^r$', branch_re='^b$'
        )
        self.assertEqual(filter.is_matched(ss), expected)

        filter = SourceStampFilter(
            project_re=re.compile('^p$'),
            codebase_re=re.compile('^c$'),
            repository_re=re.compile('^r$'),
            branch_re=re.compile('^b$'),
        )
        self.assertEqual(filter.is_matched(ss), expected)

    @parameterized.expand([
        ('match', {'project': 'p', 'codebase': 'c', 'repository': 'r', 'branch': 'b'}, True),
        (
            'not_project',
            {'project': 'p0', 'codebase': 'c', 'repository': 'r', 'branch': 'b'},
            False,
        ),
        (
            'not_codebase',
            {'project': 'p', 'codebase': 'c0', 'repository': 'r', 'branch': 'b'},
            False,
        ),
        (
            'not_repository',
            {'project': 'p', 'codebase': 'c', 'repository': 'r0', 'branch': 'b'},
            False,
        ),
        ('not_branch', {'project': 'p', 'codebase': 'c', 'repository': 'r', 'branch': 'b0'}, False),
        ('none_branch', {'project': 'p', 'codebase': 'c', 'repository': 'r', 'branch': None}, True),
    ])
    def test_filter_is_matched_not_eq_or_re(self, name, ss, expected):
        filter = SourceStampFilter(
            project_not_eq='p0', codebase_not_eq='c0', repository_not_eq='r0', branch_not_eq='b0'
        )
        self.assertEqual(filter.is_matched(ss), expected)

        filter = SourceStampFilter(
            project_not_re='^p0$',
            codebase_not_re='^c0$',
            repository_not_re='^r0$',
            branch_not_re='^b0$',
        )
        self.assertEqual(filter.is_matched(ss), expected)

        filter = SourceStampFilter(
            project_not_re=re.compile('^p0$'),
            codebase_not_re=re.compile('^c0$'),
            repository_not_re=re.compile('^r0$'),
            branch_not_re=re.compile('^b0$'),
        )
        self.assertEqual(filter.is_matched(ss), expected)

    def test_filter_repr(self):
        filter = SourceStampFilter(
            project_eq='p',
            codebase_eq='c',
            repository_eq='r',
            branch_eq='b',
            project_re='^p$',
            codebase_re='^c$',
            repository_re='^r$',
            branch_re='^b$',
            project_not_eq='p0',
            codebase_not_eq='c0',
            repository_not_eq='r0',
            branch_not_eq='b0',
            project_not_re='^p0$',
            codebase_not_re='^c0$',
            repository_not_re='^r0$',
            branch_not_re='^b0$',
        )
        self.assertEqual(
            repr(filter),
            "<SourceStampFilter on project in ['p'] and project not in ['p0'] "
            + "and project matches [re.compile('^p$')] "
            + "and project does not match [re.compile('^p0$')] "
            + "and codebase in ['c'] and codebase not in ['c0'] "
            + "and codebase matches [re.compile('^c$')] "
            + "and codebase does not match [re.compile('^c0$')] "
            + "and repository in ['r'] and repository not in ['r0'] "
            + "and repository matches [re.compile('^r$')] "
            + "and repository does not match [re.compile('^r0$')] "
            + "and branch in ['b'] and branch not in ['b0'] "
            + "and branch matches [re.compile('^b$')] "
            + "and branch does not match [re.compile('^b0$')]>",
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from buildbot.test.fakedb.row import Row


class User(Row):
    table = "users"

    id_column = 'uid'

    def __init__(self, uid=None, identifier='soap', bb_username=None, bb_password=None):
        super().__init__(
            uid=uid, identifier=identifier, bb_username=bb_username, bb_password=bb_password
        )


class UserInfo(Row):
    table = "users_info"

    def __init__(self, uid=None, attr_type='git', attr_data='Tyler Durden <tyler@mayhem.net>'):
        super().__init__(uid=uid, attr_type=attr_type, attr_data=attr_data)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import base64
import textwrap
from typing import TYPE_CHECKING
from unittest import mock

import sqlalchemy as sa
from twisted.internet import defer
from twisted.trial import unittest

from buildbot.db import compression
from buildbot.db import logs
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.util import bytes2unicode
from buildbot.util import unicode2bytes
from buildbot.util.twisted import async_to_deferred

if TYPE_CHECKING:
    from typing import Callable


class FakeUnavailableCompressor(compression.CompressorInterface):
    name = "fake"
    available = False

    HEADER = b"[FakeHeader]"

    @staticmethod
    def dumps(data: bytes) -> bytes:
        return FakeUnavailableCompressor.HEADER + data

    @staticmethod
    def read(data: bytes) -> bytes:
        assert data.startswith(FakeUnavailableCompressor.HEADER)
        return data[len(FakeUnavailableCompressor.HEADER) :]


class Tests(TestReactorMixin, unittest.TestCase):
    TIMESTAMP_STEP101 = 100000
    TIMESTAMP_STEP102 = 200000
    backgroundData = [
        fakedb.Worker(id=47, name='linux'),
        fakedb.Buildset(id=20),
        fakedb.Builder(id=88, name='b1'),
        fakedb.BuildRequest(id=41, buildsetid=20, builderid=88),
        fakedb.Master(id=88),
        fakedb.Build(id=30, buildrequestid=41, number=7, masterid=88, builderid=88, workerid=47),
        fakedb.Step(id=101, buildid=30, number=1, name='one', started_at=TIMESTAMP_STEP101),
        fakedb.Step(id=102, buildid=30, number=2, name='two', started_at=TIMESTAMP_STEP102),
    ]
    insert_test_data: Callable[[list], defer.Deferred]

    testLogLines = [
        fakedb.Log(
            id=201, stepid=101, name='stdio', slug='stdio', complete=0, num_lines=7, type='s'
        ),
        fakedb.LogChunk(
            logid=201,
            first_line=0,
            last_line=1,
            compressed=0,
            content=textwrap.dedent(
                """\
                    line zero
                    line 1"""
                + "x" * 200
            ),
        ),
        fakedb.LogChunk(
            logid=201,
            first_line=2,
            last_line=4,
            compressed=0,
            content=textwrap.dedent("""\
                    line TWO

                    line 2**2"""),
        ),
        fakedb.LogChunk(logid=201, first_line=5, last_line=5, compressed=0, content="another line"),
        fakedb.LogChunk(
            logid=201, first_line=6, last_line=6, compressed=0, content="yet another line"
        ),
    ]
    bug3101Content = base64.b64decode("""
        PT09PT09PT09PT09PT09PT09PT09PT09PT09PT09PT09PT09PT09PT09PT09PT09PT0
        9PT09PT09PT09PT09PT09PT09PT09PT09PT09PQpbU0tJUFBFRF0Kbm90IGEgd2luMz
        IgcGxhdGZvcm0KCmJ1aWxkc2xhdmUudGVzdC51bml0LnRlc3RfcnVucHJvY2Vzcy5UZ
        XN0UnVuUHJvY2Vzcy50ZXN0UGlwZVN0cmluZwotLS0tLS0tLS0tLS0tLS0tLS0tLS0t
        LS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0
        tLS0tLS0tClJhbiAyNjcgdGVzdHMgaW4gNS4zNzhzCgpQQVNTRUQgKHNraXBzPTEsIH
        N1Y2Nlc3Nlcz0yNjYpCnByb2dyYW0gZmluaXNoZWQgd2l0aCBleGl0IGNvZGUgMAplb
        GFwc2VkVGltZT04LjI0NTcwMg==""")

    bug3101Rows = [
        fakedb.Log(
            id=1470,
            stepid=101,
            name='problems',
            slug='problems',
            complete=1,
            num_lines=11,
            type='t',
        ),
        fakedb.LogChunk(
            logid=1470, first_line=0, last_line=10, compressed=0, content=bug3101Content
        ),
    ]

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantDb=True)
        self.db = self.master.db

    @defer.inlineCallbacks
    def checkTestLogLines(self):
        expLines = [
            'line zero',
            'line 1' + "x" * 200,
            'line TWO',
            '',
            'line 2**2',
            'another line',
            'yet another line',
        ]

        def _join_lines(lines: list[str]):
            return ''.join(e + '\n' for e in lines)

        for first_line in range(0, 7):
            for last_line in range(first_line, 7):
                got_lines = yield self.db.logs.getLogLines(201, first_line, last_line)
                self.assertEqual(got_lines, _join_lines(expLines[first_line : last_line + 1]))
        # check overflow
        self.assertEqual((yield self.db.logs.getLogLines(201, 5, 20)), _join_lines(expLines[5:7]))

    @defer.inlineCallbacks
    def test_getLog(self):
        yield self.db.insert_test_data([
            *self.backgroundData,
            fakedb.Log(
                id=201, stepid=101, name="stdio", slug="stdio", complete=0, num_lines=200, type="s"
            ),
        ])
        logdict = yield self.db.logs.getLog(201)
        self.assertIsInstance(logdict, logs.LogModel)
        self.assertEqual(
            logdict,
            logs.LogModel(
                id=201,
                stepid=101,
                name='stdio',
                slug='stdio',
                complete=False,
                num_lines=200,
                type='s',
            ),
        )

    @defer.inlineCallbacks
    def test_getLog_missing(self):
        logdict = yield self.db.logs.getLog(201)
        self.assertEqual(logdict, None)

    @defer.inlineCallbacks
    def test_getLogBySlug(self):
        yield self.db.insert_test_data([
            *self.backgroundData,
            fakedb.Log(
                id=201, stepid=101, name="stdio", slug="stdio", complete=0, num_lines=200, type="s"
            ),
            fakedb.Log(
                id=202,
                stepid=101,
                name="dbg.log",
                slug="dbg_log",
                complete=1,
                num_lines=200,
                type="s",
            ),
        ])
        logdict = yield self.db.logs.getLogBySlug(101, 'dbg_log')
        self.assertIsInstance(logdict, logs.LogModel)
        self.assertEqual(logdict.id, 202)

    @defer.inlineCallbacks
    def test_getLogBySlug_missing(self):
        yield self.db.insert_test_data([
            *self.backgroundData,
            fakedb.Log(
                id=201, stepid=101, name="stdio", slug="stdio", complete=0, num_lines=200, type="s"
            ),
        ])
        logdict = yield self.db.logs.getLogBySlug(102, 'stdio')
        self.assertEqual(logdict, None)

    @defer.inlineCallbacks
    def test_getLogs(self):
        yield self.db.insert_test_data([
            *self.backgroundData,
            fakedb.Log(
                id=201, stepid=101, name="stdio", slug="stdio", complete=0, num_lines=200, type="s"
            ),
            fakedb.Log(
                id=202,
                stepid=101,
                name="dbg.log",
                slug="dbg_log",
                complete=1,
                num_lines=300,
                type="t",
            ),
            fakedb.Log(
                id=203, stepid=102, name="stdio", slug="stdio", complete=0, num_lines=200, type="s"
            ),
        ])
        logdicts = yield self.db.logs.getLogs(101)
        for logdict in logdicts:
            self.assertIsInstance(logdict, logs.LogModel)
        self.assertEqual(sorted([ld.id for ld in logdicts]), [201, 202])

    @defer.inlineCallbacks
    def test_getLogLines(self):
        yield self.db.insert_test_data(self.backgroundData + self.testLogLines)
        yield self.checkTestLogLines()

        # check line number reversal
        self.assertEqual((yield self.db.logs.getLogLines(201, 6, 3)), '')

    @defer.inlineCallbacks
    def test_getLogLines_empty(self):
        yield self.db.insert_test_data([
            *self.backgroundData,
            fakedb.Log(
                id=201, stepid=101, name="stdio", slug="stdio", complete=0, num_lines=200, type="s"
            ),
        ])
        self.assertEqual((yield self.db.logs.getLogLines(201, 9, 99)), '')
        self.assertEqual((yield self.db.logs.getLogLines(999, 9, 99)), '')

    @defer.inlineCallbacks
    def test_getLogLines_bug3101(self):
        # regression test for #3101
        content = self.bug3101Content
        yield self.db.insert_test_data(self.backgroundData + self.bug3101Rows)
        # overall content is the same, with '\n' padding at the end
        expected = bytes2unicode(self.bug3101Content + b'\n')
        self.assertEqual((yield self.db.logs.getLogLines(1470, 0, 99)), expected)
        # try to fetch just one line
        expected = bytes2unicode(content.split(b'\n')[0] + b'\n')
        self.assertEqual((yield self.db.logs.getLogLines(1470, 0, 0)), expected)

    @defer.inlineCallbacks
    def test_addLog_getLog(self):
        yield self.db.insert_test_data(self.backgroundData)
        logid = yield self.db.logs.addLog(
            stepid=101, name='config.log', slug='config_log', type='t'
        )
        logdict = yield self.db.logs.getLog(logid)
        self.assertIsInstance(logdict, logs.LogModel)
        self.assertEqual(
            logdict,
            logs.LogModel(
                id=logid,
                stepid=101,
                name='config.log',
                slug='config_log',
                complete=False,
                num_lines=0,
                type='t',
            ),
        )

    @defer.inlineCallbacks
    def test_appendLog_getLogLines(self):
        yield self.db.insert_test_data(self.backgroundData + self.testLogLines)
        logid = yield self.db.logs.addLog(stepid=102, name='another', slug='another', type='s')
        self.assertEqual((yield self.db.logs.appendLog(logid, 'xyz\n')), (0, 0))
        self.assertEqual((yield self.db.logs.appendLog(201, 'abc\ndef\n')), (7, 8))
        self.assertEqual((yield self.db.logs.appendLog(logid, 'XYZ\n')), (1, 1))
        self.assertEqual((yield self.db.logs.getLogLines(201, 6, 7)), "yet another line\nabc\n")
        self.assertEqual((yield self.db.logs.getLogLines(201, 7, 8)), "abc\ndef\n")
        self.assertEqual((yield self.db.logs.getLogLines(201, 8, 8)), "def\n")
        self.assertEqual((yield self.db.logs.getLogLines(logid, 0, 1)), "xyz\nXYZ\n")
        self.assertEqual(
            (yield self.db.logs.getLog(logid)),
            logs.LogModel(
                complete=False,
                id=logid,
                name='another',
                slug='another',
                num_lines=2,
                stepid=102,
                type='s',
            ),
        )

    @defer.inlineCallbacks
    def test_compressLog(self):
        yield self.db.insert_test_data(self.backgroundData + self.testLogLines)
        yield self.db.logs.compressLog(201)
        # test log lines should still be readable just the same
        yield self.checkTestLogLines()

    @defer.inlineCallbacks
    def test_addLogLines_big_chunk(self):
        yield self.db.insert_test_data(self.backgroundData + self.testLogLines)
        self.assertEqual(
            (yield self.db.logs.appendLog(201, 'abc\n' * 20000)),  # 80k
            (7, 20006),
        )
        lines = yield self.db.logs.getLogLines(201, 7, 50000)
        self.assertEqual(len(lines), 80000)
        self.assertEqual(lines, ('abc\n' * 20000))

    @defer.inlineCallbacks
    def test_addLogLines_big_chunk_big_lines(self):
        yield self.db.insert_test_data(self.backgroundData + self.testLogLines)
        line = 'x' * 33000 + '\n'
        self.assertEqual(
            (yield self.db.logs.appendLog(201, line * 3)), (7, 9)
        )  # three long lines, all truncated
        lines = yield self.db.logs.getLogLines(201, 7, 100)
        self.assertEqual(len(lines), 99003)
        self.assertEqual(lines, (line * 3))

    @defer.inlineCallbacks
    def test_addLogLines_db(self):
        yield self.db.insert_test_data(self.backgroundData + self.testLogLines)
        self.assertEqual((yield self.db.logs.appendLog(201, 'abc\ndef\nghi\njkl\n')), (7, 10))

        def thd(conn):
            res = conn.execute(
                self.db.model.logchunks.select().where(self.db.model.logchunks.c.first_line > 6)
            ).mappings()
            row = res.fetchone()
            res.close()
            return dict(row)

        newRow = yield self.db.pool.do(thd)
        self.assertEqual(
            newRow,
            {
                'logid': 201,
                'first_line': 7,
                'last_line': 10,
                'content': b'abc\ndef\nghi\njkl',
                'compressed': 0,
            },
        )

    async def _base_appendLog_truncate(self, content: str):
        LOG_ID = 201
        await self.db.insert_test_data([
            *self.backgroundData,
            fakedb.Log(
                id=LOG_ID,
                stepid=101,
                name='stdio',
                slug='stdio',
                complete=0,
                num_lines=0,
                type='s',
            ),
        ])
        await self.db.logs.appendLog(LOG_ID, content)

        def _thd(conn: sa.engine.Connection) -> list[dict]:
            tbl = self.db.model.logchunks
            res = conn.execute(
                tbl.select().where(tbl.c.logid == LOG_ID).order_by(tbl.c.first_line)
            ).mappings()
            rows = [dict(row) for row in res]
            res.close()
            return rows

        return await self.db.pool.do(_thd)

    @async_to_deferred
    async def test_appendLog_no_truncate_compressable_chunks(self):
        content = 'a ' + '\N{SNOWMAN}' * 100000 + '\n'
        assert len(content) > self.db.logs.MAX_CHUNK_SIZE
        self.db.master.config.logCompressionMethod = "gz"
        rows = await self._base_appendLog_truncate(content)
        self.assertEqual(
            [
                {
                    'compressed': 1,
                    'content': self.db.logs._get_compressor(1).dumps(content[:-1].encode('utf-8')),
                    'first_line': 0,
                    'last_line': 0,
                    'logid': 201,
                }
            ],
            rows,
        )

    @async_to_deferred
    async def test_appendLog_truncate_chunk(self):
        self.maxDiff = None
        content = 'a ' + '\N{SNOWMAN}' * 100000 + '\n'
        assert len(content) > self.db.logs.MAX_CHUNK_SIZE
        self.db.master.config.logCompressionMethod = "raw"
        rows = await self._base_appendLog_truncate(content)
        self.assertTrue(len(rows[0].pop('content')) <= self.db.logs.MAX_CHUNK_SIZE)
        self.assertEqual(
            [
                {
                    'compressed': 0,
                    'first_line': 0,
                    'last_line': 0,
                    'logid': 201,
                }
            ],
            rows,
        )

    @defer.inlineCallbacks
    def test_no_compress_small_chunk(self):
        yield self.db.insert_test_data(self.backgroundData + self.testLogLines)
        self.db.master.config.logCompressionMethod = "gz"
        self.assertEqual((yield self.db.logs.appendLog(201, 'abc\n')), (7, 7))

        def thd(conn):
            res = conn.execute(
                self.db.model.logchunks.select().where(self.db.model.logchunks.c.first_line > 6)
            ).mappings()
            row = res.fetchone()
            res.close()
            return dict(row)

        newRow = yield self.db.pool.do(thd)
        self.assertEqual(
            newRow,
            {'logid': 201, 'first_line': 7, 'last_line': 7, 'content': b'abc', 'compressed': 0},
        )

    async def _test_compress_big_chunk(
        self,
        compressor: compression.CompressorInterface,
        compressed_id: int,
    ) -> None:
        await self.db.insert_test_data(self.backgroundData + self.testLogLines)
        line = 'xy' * 10000
        self.assertEqual((await self.db.logs.appendLog(201, line + '\n')), (7, 7))

        def thd(conn):
            res = conn.execute(
                self.db.model.logchunks.select().where(self.db.model.logchunks.c.first_line > 6)
            ).mappings()
            row = res.fetchone()
            res.close()
            return dict(row)

        newRow = await self.db.pool.do(thd)
        self.assertEqual(compressor.read(newRow.pop('content')), unicode2bytes(line))
        self.assertEqual(
            newRow,
            {
                'logid': 201,
                'first_line': 7,
                'last_line': 7,
                'compressed': compressed_id,
            },
        )

    @async_to_deferred
    async def test_raw_compress_big_chunk(self):
        fake_raw_compressor = mock.Mock(spec=compression.CompressorInterface)
        fake_raw_compressor.read = lambda d: d
        self.db.master.config.logCompressionMethod = "raw"
        await self._test_compress_big_chunk(fake_raw_compressor, 0)

    @async_to_deferred
    async def test_gz_compress_big_chunk(self):
        self.db.master.config.logCompressionMethod = "gz"
        await self._test_compress_big_chunk(compression.GZipCompressor, 1)

    @async_to_deferred
    async def test_bz2_compress_big_chunk(self):
        self.db.master.config.logCompressionMethod = "bz2"
        await self._test_compress_big_chunk(compression.BZipCompressor, 2)

    @async_to_deferred
    async def test_lz4_compress_big_chunk(self):
        try:
            import lz4  # noqa: F401
        except ImportError as e:
            raise unittest.SkipTest("lz4 not installed, skip the test") from e

        self.db.master.config.logCompressionMethod = "lz4"
        await self._test_compress_big_chunk(compression.LZ4Compressor, 3)

    @async_to_deferred
    async def test_zstd_compress_big_chunk(self):
        try:
            import zstandard  # noqa: F401
        except ImportError as e:
            raise unittest.SkipTest("zstandard not installed, skip the test") from e

        self.db.master.config.logCompressionMethod = "zstd"
        await self._test_compress_big_chunk(compression.ZStdCompressor, 4)

    @async_to_deferred
    async def test_br_compress_big_chunk(self):
        try:
            import brotli  # noqa: F401
        except ImportError as e:
            raise unittest.SkipTest("brotli not installed, skip the test") from e

        self.db.master.config.logCompressionMethod = "br"
        await self._test_compress_big_chunk(compression.BrotliCompressor, 5)

    @defer.inlineCallbacks
    def do_addLogLines_huge_log(self, NUM_CHUNKS=3000, chunk=('xy' * 70 + '\n') * 3):
        if chunk.endswith("\n"):
            chunk = chunk[:-1]
        linesperchunk = chunk.count("\n") + 1
        test_data = [
            fakedb.LogChunk(
                logid=201,
                first_line=i * linesperchunk,
                last_line=i * linesperchunk + linesperchunk - 1,
                compressed=0,
                content=chunk,
            )
            for i in range(NUM_CHUNKS)
        ]
        yield self.db.insert_test_data([
            *self.backgroundData,
            fakedb.Log(
                id=201,
                stepid=101,
                name="stdio",
                slug="stdio",
                complete=0,
                num_lines=NUM_CHUNKS * 3,
                type="s",
            ),
            *test_data,
        ])
        wholeLog = yield self.db.logs.getLogLines(201, 0, NUM_CHUNKS * 3)
        for i in range(10):
            yield self.db.logs.compressLog(201)
            wholeLog2 = yield self.db.logs.getLogLines(201, 0, NUM_CHUNKS * 3)
            self.assertEqual(wholeLog, wholeLog2)
        self.assertEqual(wholeLog, wholeLog2)

        def countChunk(conn):
            tbl = self.db.model.logchunks
            q = sa.select(sa.func.count(tbl.c.content))
            q = q.where(tbl.c.logid == 201)
            return conn.execute(q).fetchone()[0]

        chunks = yield self.db.pool.do(countChunk)
        # make sure MAX_CHUNK_LINES is taken in account
        self.assertGreaterEqual(
            chunks, NUM_CHUNKS * linesperchunk / logs.LogsConnectorComponent.MAX_CHUNK_LINES
        )

    def test_addLogLines_huge_log(self):
        return self.do_addLogLines_huge_log()

    def test_addLogLines_huge_log_lots_line(self):
        return self.do_addLogLines_huge_log(NUM_CHUNKS=3000, chunk='x\n' * 50)

    def test_addLogLines_huge_log_lots_snowmans(self):
        return self.do_addLogLines_huge_log(NUM_CHUNKS=3000, chunk='\N{SNOWMAN}\n' * 50)

    @defer.inlineCallbacks
    def test_compressLog_non_existing_log(self):
        yield self.db.logs.compressLog(201)
        logdict = yield self.db.logs.getLog(201)
        self.assertEqual(logdict, None)

    @defer.inlineCallbacks
    def test_compressLog_empty_log(self):
        yield self.db.insert_test_data([
            *self.backgroundData,
            fakedb.Log(
                id=201, stepid=101, name="stdio", slug="stdio", complete=1, num_lines=0, type="s"
            ),
        ])
        yield self.db.logs.compressLog(201)
        logdict = yield self.db.logs.getLog(201)
        self.assertEqual(
            logdict,
            logs.LogModel(
                stepid=101,
                num_lines=0,
                name='stdio',
                id=201,
                type='s',
                slug='stdio',
                complete=True,
            ),
        )

    @defer.inlineCallbacks
    def test_deleteOldLogChunks_basic(self):
        yield self.db.insert_test_data(self.backgroundData)
        logids = []
        for stepid in (101, 102):
            for i in range(stepid):
                logid = yield self.db.logs.addLog(
                    stepid=stepid, name='another' + str(i), slug='another' + str(i), type='s'
                )
                yield self.db.logs.appendLog(logid, 'xyz\n')
                logids.append(logid)

        deleted_chunks = yield self.db.logs.deleteOldLogChunks(
            (self.TIMESTAMP_STEP102 + self.TIMESTAMP_STEP101) / 2
        )
        self.assertEqual(deleted_chunks, 101)
        deleted_chunks = yield self.db.logs.deleteOldLogChunks(
            self.TIMESTAMP_STEP102 + self.TIMESTAMP_STEP101
        )
        self.assertEqual(deleted_chunks, 102)
        deleted_chunks = yield self.db.logs.deleteOldLogChunks(
            self.TIMESTAMP_STEP102 + self.TIMESTAMP_STEP101
        )
        self.assertEqual(deleted_chunks, 0)
        deleted_chunks = yield self.db.logs.deleteOldLogChunks(0)
        self.assertEqual(deleted_chunks, 0)
        for logid in logids:
            logdict = yield self.db.logs.getLog(logid)
            self.assertEqual(logdict.type, 'd')

            # we make sure we can still getLogLines, it will just return empty value
            lines = yield self.db.logs.getLogLines(logid, 0, logdict.num_lines)
            self.assertEqual(lines, '')

    @async_to_deferred
    async def test_insert_logs_non_existing_compression_method(self):
        LOG_ID = 201
        await self.db.insert_test_data([
            *self.backgroundData,
            fakedb.Log(
                id=LOG_ID, stepid=101, name="stdio", slug="stdio", complete=0, num_lines=1, type="s"
            ),
            fakedb.LogChunk(
                logid=LOG_ID, first_line=0, last_line=0, compressed=0, content=b"fake_log_chunk\n"
            ),
        ])

        def _thd_get_log_chunks(conn):
            res = conn.execute(
                self.db.model.logchunks.select().where(self.db.model.logchunks.c.logid == LOG_ID)
            ).mappings()
            return [dict(row) for row in res]

        self.db.master.config.logCompressionMethod = "non_existing"
        await self.db.logs.compressLog(LOG_ID)

        self.assertEqual(
            await self.db.pool.do(_thd_get_log_chunks),
            [
                {
                    'compressed': 0,
                    'content': b'fake_log_chunk\n',
                    'first_line': 0,
                    'last_line': 0,
                    'logid': LOG_ID,
                }
            ],
        )

        await self.db.logs.appendLog(LOG_ID, 'other_chunk\n')

        self.assertEqual(
            await self.db.pool.do(_thd_get_log_chunks),
            [
                {
                    'compressed': 0,
                    'content': b'fake_log_chunk\n',
                    'first_line': 0,
                    'last_line': 0,
                    'logid': LOG_ID,
                },
                {
                    'compressed': 0,
                    'content': b'other_chunk',
                    'first_line': 1,
                    'last_line': 1,
                    'logid': LOG_ID,
                },
            ],
        )

    @async_to_deferred
    async def test_get_logs_non_existing_compression_method(self):
        LOG_ID = 201

        # register fake compressor
        FAKE_COMPRESSOR_ID = max(self.db.logs.COMPRESSION_BYID.keys()) + 1
        self.db.logs.COMPRESSION_BYID[FAKE_COMPRESSOR_ID] = FakeUnavailableCompressor
        NON_EXISTING_COMPRESSOR_ID = max(self.db.logs.COMPRESSION_BYID.keys()) + 1

        await self.db.insert_test_data([
            *self.backgroundData,
            fakedb.Log(
                id=LOG_ID, stepid=101, name="stdio", slug="stdio", complete=0, num_lines=1, type="s"
            ),
            fakedb.LogChunk(
                logid=LOG_ID,
                first_line=0,
                last_line=0,
                compressed=FAKE_COMPRESSOR_ID,
                content=b"fake_log_chunk\n",
            ),
            fakedb.LogChunk(
                logid=LOG_ID,
                first_line=1,
                last_line=1,
                compressed=NON_EXISTING_COMPRESSOR_ID,
                content=b"fake_log_chunk\n",
            ),
        ])

        with self.assertRaises(logs.LogCompressionFormatUnavailableError):
            await self.db.logs.getLogLines(logid=LOG_ID, first_line=0, last_line=0)

        with self.assertRaises(logs.LogCompressionFormatUnavailableError):
            await self.db.logs.getLogLines(logid=LOG_ID, first_line=1, last_line=1)
        self.flushLoggedErrors(logs.LogCompressionFormatUnavailableError)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import operator
import posixpath
from pathlib import PurePosixPath
from unittest.mock import Mock
from unittest.mock import call

from twisted.internet import defer
from twisted.trial import unittest
from zope.interface import implementer

from buildbot import interfaces
from buildbot.locks import WorkerLock
from buildbot.process.build import Build
from buildbot.process.buildstep import BuildStep
from buildbot.process.buildstep import create_step_from_step_or_factory
from buildbot.process.locks import get_real_locks_from_accesses
from buildbot.process.metrics import MetricLogObserver
from buildbot.process.properties import Properties
from buildbot.process.results import CANCELLED
from buildbot.process.results import EXCEPTION
from buildbot.process.results import FAILURE
from buildbot.process.results import RETRY
from buildbot.process.results import SUCCESS
from buildbot.process.results import WARNINGS
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.fake import fakeprotocol
from buildbot.test.fake import worker
from buildbot.test.reactor import TestReactorMixin


class FakeChange:
    def __init__(self, number=None):
        self.properties = Properties()
        self.number = number
        self.who = "me"


class FakeSource:
    def __init__(self):
        self.sourcestampsetid = None
        self.changes = []
        self.branch = None
        self.revision = None
        self.repository = ''
        self.codebase = ''
        self.project = ''
        self.patch_info = None
        self.patch = None

    def getRepository(self):
        return self.repository


class FakeRequest:
    def __init__(self):
        self.sources = []
        self.reason = "Because"
        self.properties = Properties()
        self.id = 9385

    def mergeSourceStampsWith(self, others):
        return self.sources

    def mergeReasons(self, others):
        return self.reason


class FakeBuildStep(BuildStep):
    def __init__(self):
        super().__init__(
            haltOnFailure=False,
            flunkOnWarnings=False,
            flunkOnFailure=True,
            warnOnWarnings=True,
            warnOnFailure=False,
            alwaysRun=False,
            name='fake',
        )
        self._summary = {'step': 'result', 'build': 'build result'}
        self._expected_results = SUCCESS

    def run(self):
        return self._expected_results

    def getResultSummary(self):
        return self._summary

    def interrupt(self, reason):
        self.running = False
        self.interrupted = reason


class FakeBuilder:
    def __init__(self, master):
        self.config = Mock()
        self.config.workerbuilddir = 'wbd'
        self.config.description = 'builder-description'
        self.config.env = {}
        self.name = 'fred'
        self.master = master
        self.botmaster = master.botmaster
        self.builderid = 83
        self._builders = {}
        self.config_version = 0

    def getBuilderId(self):
        return defer.succeed(self.builderid)

    def setup_properties(self, props):
        return defer.succeed(None)

    def buildFinished(self, build, workerforbuilder):
        pass

    def getBuilderIdForName(self, name):
        return defer.succeed(self._builders.get(name, None) or self.builderid)

    def find_project_id(self, name):
        return defer.succeed(None)


@implementer(interfaces.IBuildStepFactory)
class FakeStepFactory:
    """Fake step factory that just returns a fixed step object."""

    def __init__(self, step):
        self.step = step

    def buildStep(self):
        return self.step


class TestException(Exception):
    pass


@implementer(interfaces.IBuildStepFactory)
class FailingStepFactory:
    """Fake step factory that just returns a fixed step object."""

    def buildStep(self):
        raise TestException("FailingStepFactory")


class _StepController:
    def __init__(self, step):
        self._step = step

    def finishStep(self, result):
        self._step._deferred.callback(result)


class _ControllableStep(BuildStep):
    def __init__(self):
        super().__init__()
        self._deferred = defer.Deferred()

    def run(self):
        return self._deferred


def makeControllableStepFactory():
    step = create_step_from_step_or_factory(_ControllableStep())
    controller = _StepController(step)
    return controller, FakeStepFactory(step)


class TestBuild(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        r = FakeRequest()
        r.sources = [FakeSource()]
        r.sources[0].changes = [FakeChange()]
        r.sources[0].revision = "12345"

        self.request = r
        self.master = yield fakemaster.make_master(self, wantData=True)

        yield self.master.db.insert_test_data([
            fakedb.Master(id=fakedb.FakeDBConnector.MASTER_ID),
            fakedb.Worker(id=1234),
            fakedb.Builder(id=83),
            fakedb.Buildset(id=8822),
            fakedb.BuildRequest(id=9385, builderid=83, buildsetid=8822),
        ])

        self.worker = worker.FakeWorker(self.master)
        self.worker.attached(None)
        self.builder = FakeBuilder(self.master)
        self.build = Build([r], self.builder)
        self.build.conn = fakeprotocol.FakeConnection(self.worker)
        self.build.workername = self.worker.workername

        self.workerforbuilder = Mock(name='workerforbuilder')
        self.workerforbuilder.worker = self.worker
        self.workerforbuilder.substantiate_if_needed = lambda _: True
        self.workerforbuilder.ping = lambda: True

        self.build.workerforbuilder = self.workerforbuilder
        self.build.text = []
        self.build.buildid = 666

    @defer.inlineCallbacks
    def assert_worker_preparation_failure(self, reason):
        steps = yield self.master.data.get(('builds', self.build.buildid, 'steps'))
        self.assertIn(steps[-1]['state_string'], reason)

    def create_fake_build_step(self):
        return create_step_from_step_or_factory(FakeBuildStep())

    def _setup_lock_claim_log(self, lock, claim_log):
        if hasattr(lock, "_old_claim"):
            return

        def claim(owner, access):
            claim_log.append(owner)
            return lock._old_claim(owner, access)

        lock._old_claim = lock.claim
        lock.claim = claim

    def testRunSuccessfulBuild(self):
        b = self.build

        step = self.create_fake_build_step()
        b.setStepFactories([FakeStepFactory(step)])

        b.startBuild(self.workerforbuilder)

        self.assertEqual(b.results, SUCCESS)

    def testStopBuild(self):
        b = self.build

        step = self.create_fake_build_step()
        b.setStepFactories([FakeStepFactory(step)])

        def startStep(*args, **kw):
            # Now interrupt the build
            b.stopBuild("stop it")
            return defer.Deferred()

        step.startStep = startStep

        b.startBuild(self.workerforbuilder)

        self.assertEqual(b.results, CANCELLED)

        self.assertIn('stop it', step.interrupted)

    @defer.inlineCallbacks
    def test_build_retry_when_worker_substantiate_returns_false(self):
        b = self.build

        step = self.create_fake_build_step()
        b.setStepFactories([FakeStepFactory(step)])

        self.workerforbuilder.substantiate_if_needed = lambda _: False

        yield b.startBuild(self.workerforbuilder)
        self.assertEqual(b.results, RETRY)
        yield self.assert_worker_preparation_failure('error while worker_prepare')

    @defer.inlineCallbacks
    def test_build_cancelled_when_worker_substantiate_returns_false_due_to_cancel(self):
        b = self.build

        step = self.create_fake_build_step()
        b.setStepFactories([FakeStepFactory(step)])

        substantiation_d = defer.Deferred()
        self.workerforbuilder.substantiate_if_needed = lambda _: substantiation_d

        build_d = b.startBuild(self.workerforbuilder)
        b.stopBuild('Cancel Build', CANCELLED)

        substantiation_d.callback(False)
        yield build_d

        self.assertEqual(b.results, CANCELLED)
        yield self.assert_worker_preparation_failure('pending')

    @defer.inlineCallbacks
    def test_build_retry_when_worker_substantiate_returns_false_due_to_cancel(self):
        b = self.build

        step = self.create_fake_build_step()
        b.setStepFactories([FakeStepFactory(step)])

        d = defer.Deferred()
        self.workerforbuilder.substantiate_if_needed = lambda _: d
        b.startBuild(self.workerforbuilder)
        b.stopBuild('Cancel Build', RETRY)
        d.callback(False)
        self.assertEqual(b.results, RETRY)
        yield self.assert_worker_preparation_failure('pending')

    @defer.inlineCallbacks
    def testAlwaysRunStepStopBuild(self):
        """Test that steps marked with alwaysRun=True still get run even if
        the build is stopped."""

        # Create a build with 2 steps, the first one will get interrupted, and
        # the second one is marked with alwaysRun=True
        b = self.build

        step1 = self.create_fake_build_step()
        step1.alwaysRun = False
        step1.results = None
        step2 = self.create_fake_build_step()
        step2.alwaysRun = True
        step2.results = None
        b.setStepFactories([
            FakeStepFactory(step1),
            FakeStepFactory(step2),
        ])

        def startStep1(*args, **kw):
            # Now interrupt the build
            b.stopBuild("stop it")
            return defer.succeed(SUCCESS)

        step1.startStep = startStep1
        step1.stepDone = lambda: False

        step2Started = [False]

        def startStep2(*args, **kw):
            step2Started[0] = True
            return defer.succeed(SUCCESS)

        step2.startStep = startStep2
        step1.stepDone = lambda: False

        yield b.startBuild(self.workerforbuilder)

        self.assertEqual(b.results, CANCELLED)
        self.assertIn('stop it', step1.interrupted)
        self.assertTrue(step2Started[0])

    @defer.inlineCallbacks
    def test_start_step_throws_exception(self):
        b = self.build

        step1 = self.create_fake_build_step()
        b.setStepFactories([
            FakeStepFactory(step1),
        ])

        def startStep(*args, **kw):
            raise TestException()

        step1.startStep = startStep

        yield b.startBuild(self.workerforbuilder)

        self.assertEqual(b.results, EXCEPTION)
        self.flushLoggedErrors(TestException)

    def testBuilddirPropType(self):
        b = self.build

        b.builder.config.workerbuilddir = 'test'
        self.workerforbuilder.worker.worker_basedir = "/srv/buildbot/worker"
        self.workerforbuilder.worker.path_module = posixpath
        self.workerforbuilder.worker.path_cls = PurePosixPath
        b.getProperties = Mock()
        b.setProperty = Mock()

        b.setupWorkerProperties(self.workerforbuilder)

        expected_path = '/srv/buildbot/worker/test'

        b.setProperty.assert_has_calls([call('builddir', expected_path, 'Worker')], any_order=True)

    @defer.inlineCallbacks
    def testBuildLocksAcquired(self):
        b = self.build

        lock = WorkerLock('lock')
        claim_log = []
        lock_access = lock.access('counting')
        lock.access = lambda mode: lock_access

        b.setLocks([lock_access])
        yield b._setup_locks()

        self._setup_lock_claim_log(b._locks_to_acquire[0][0], claim_log)

        step = self.create_fake_build_step()
        b.setStepFactories([FakeStepFactory(step)])

        b.startBuild(self.workerforbuilder)

        self.assertEqual(b.results, SUCCESS)
        self.assertEqual(len(claim_log), 1)

    @defer.inlineCallbacks
    def testBuildLocksOrder(self):
        """Test that locks are acquired in FIFO order; specifically that
        counting locks cannot jump ahead of exclusive locks"""
        eBuild = self.build
        cBuilder = FakeBuilder(self.master)
        cBuild = Build([self.request], cBuilder)
        cBuild.workerforbuilder = self.workerforbuilder
        cBuild.workername = self.worker.workername

        eWorker = Mock()
        cWorker = Mock()

        eWorker.worker = self.worker
        cWorker.worker = self.worker
        eWorker.substantiate_if_needed = cWorker.substantiate_if_needed = lambda _: True
        eWorker.ping = cWorker.ping = lambda: True

        lock = WorkerLock('lock', 2)
        claim_log = []

        eBuild.setLocks([lock.access('exclusive')])
        yield eBuild._setup_locks()

        cBuild.setLocks([lock.access('counting')])
        yield cBuild._setup_locks()

        self._setup_lock_claim_log(eBuild._locks_to_acquire[0][0], claim_log)
        self._setup_lock_claim_log(cBuild._locks_to_acquire[0][0], claim_log)

        real_lock = eBuild._locks_to_acquire[0][0]

        b3 = Mock()
        b3_access = lock.access('counting')
        real_lock.claim(b3, b3_access)

        step = self.create_fake_build_step()
        eBuild.setStepFactories([FakeStepFactory(step)])
        cBuild.setStepFactories([FakeStepFactory(step)])

        e = eBuild.startBuild(eWorker)
        c = cBuild.startBuild(cWorker)
        d = defer.DeferredList([e, c], consumeErrors=True)

        real_lock.release(b3, b3_access)

        yield d
        self.assertEqual(eBuild.results, SUCCESS)
        self.assertEqual(cBuild.results, SUCCESS)
        self.assertEqual(claim_log, [b3, eBuild, cBuild])

    @defer.inlineCallbacks
    def testBuildWaitingForLocks(self):
        b = self.build

        claim_log = []

        lock = WorkerLock('lock')
        lock_access = lock.access('counting')

        b.setLocks([lock_access])
        yield b._setup_locks()
        self._setup_lock_claim_log(b._locks_to_acquire[0][0], claim_log)

        step = self.create_fake_build_step()
        b.setStepFactories([FakeStepFactory(step)])

        real_lock = b._locks_to_acquire[0][0]
        real_lock.claim(Mock(), lock.access('counting'))

        b.startBuild(self.workerforbuilder)

        self.assertEqual(len(claim_log), 1)
        self.assertTrue(b.currentStep is None)
        self.assertTrue(b._acquiringLock is not None)

    @defer.inlineCallbacks
    def testStopBuildWaitingForLocks(self):
        b = self.build

        lock = WorkerLock('lock')
        lock_access = lock.access('counting')

        b.setLocks([lock_access])
        yield b._setup_locks()

        step = self.create_fake_build_step()
        step.alwaysRun = False
        b.setStepFactories([FakeStepFactory(step)])

        real_lock = b._locks_to_acquire[0][0]
        real_lock.claim(Mock(), lock.access('counting'))

        def acquireLocks(res=None):
            retval = Build.acquireLocks(b, res)
            b.stopBuild('stop it')
            return retval

        b.acquireLocks = acquireLocks

        b.startBuild(self.workerforbuilder)

        self.assertTrue(b.currentStep is None)
        self.assertEqual(b.results, CANCELLED)

    @defer.inlineCallbacks
    def testStopBuildWaitingForLocks_lostRemote(self):
        b = self.build

        lock = WorkerLock('lock')
        lock_access = lock.access('counting')
        lock.access = lambda mode: lock_access

        b.setLocks([lock_access])
        yield b._setup_locks()

        step = self.create_fake_build_step()
        step.alwaysRun = False
        b.setStepFactories([FakeStepFactory(step)])

        real_lock = b._locks_to_acquire[0][0]
        real_lock.claim(Mock(), lock.access('counting'))

        def acquireLocks(res=None):
            retval = Build.acquireLocks(b, res)
            b.lostRemote()
            return retval

        b.acquireLocks = acquireLocks

        b.startBuild(self.workerforbuilder)

        self.assertTrue(b.currentStep is None)
        self.assertEqual(b.results, RETRY)

    @defer.inlineCallbacks
    def testStopBuildWaitingForStepLocks(self):
        b = self.build

        lock = WorkerLock('lock')
        lock_access = lock.access('counting')

        locks = yield get_real_locks_from_accesses([lock_access], b)

        step = create_step_from_step_or_factory(BuildStep(locks=[lock_access]))
        b.setStepFactories([FakeStepFactory(step)])

        locks[0][0].claim(Mock(), lock.access('counting'))

        gotLocks = [False]

        def acquireLocks(res=None):
            gotLocks[0] = True
            retval = BuildStep.acquireLocks(step, res)
            self.assertTrue(b.currentStep is step)
            b.stopBuild('stop it')
            return retval

        step.acquireLocks = acquireLocks

        b.startBuild(self.workerforbuilder)

        self.assertEqual(gotLocks, [True])
        self.assertEqual(b.results, CANCELLED)

    def testStepDone(self):
        b = self.build
        b.results = SUCCESS
        step = self.create_fake_build_step()
        terminate = b.stepDone(SUCCESS, step)
        self.assertFalse(terminate.result)
        self.assertEqual(b.results, SUCCESS)

    def testStepDoneHaltOnFailure(self):
        b = self.build
        b.results = SUCCESS
        step = self.create_fake_build_step()
        step.haltOnFailure = True
        terminate = b.stepDone(FAILURE, step)
        self.assertTrue(terminate.result)
        self.assertEqual(b.results, FAILURE)

    def testStepDoneHaltOnFailureNoFlunkOnFailure(self):
        b = self.build
        b.results = SUCCESS
        step = self.create_fake_build_step()
        step.flunkOnFailure = False
        step.haltOnFailure = True
        terminate = b.stepDone(FAILURE, step)
        self.assertTrue(terminate.result)
        self.assertEqual(b.results, SUCCESS)

    def testStepDoneFlunkOnWarningsFlunkOnFailure(self):
        b = self.build
        b.results = SUCCESS
        step = self.create_fake_build_step()
        step.flunkOnFailure = True
        step.flunkOnWarnings = True
        b.stepDone(WARNINGS, step)
        terminate = b.stepDone(FAILURE, step)
        self.assertFalse(terminate.result)
        self.assertEqual(b.results, FAILURE)

    def testStepDoneNoWarnOnWarnings(self):
        b = self.build
        b.results = SUCCESS
        step = self.create_fake_build_step()
        step.warnOnWarnings = False
        terminate = b.stepDone(WARNINGS, step)
        self.assertFalse(terminate.result)
        self.assertEqual(b.results, SUCCESS)

    def testStepDoneWarnings(self):
        b = self.build
        b.results = SUCCESS
        step = self.create_fake_build_step()
        terminate = b.stepDone(WARNINGS, step)
        self.assertFalse(terminate.result)
        self.assertEqual(b.results, WARNINGS)

    def testStepDoneFail(self):
        b = self.build
        b.results = SUCCESS
        step = self.create_fake_build_step()
        terminate = b.stepDone(FAILURE, step)
        self.assertFalse(terminate.result)
        self.assertEqual(b.results, FAILURE)

    def testStepDoneFailOverridesWarnings(self):
        b = self.build
        b.results = WARNINGS
        step = self.create_fake_build_step()
        terminate = b.stepDone(FAILURE, step)
        self.assertFalse(terminate.result)
        self.assertEqual(b.results, FAILURE)

    def testStepDoneWarnOnFailure(self):
        b = self.build
        b.results = SUCCESS
        step = self.create_fake_build_step()
        step.warnOnFailure = True
        step.flunkOnFailure = False
        terminate = b.stepDone(FAILURE, step)
        self.assertFalse(terminate.result)
        self.assertEqual(b.results, WARNINGS)

    def testStepDoneFlunkOnWarnings(self):
        b = self.build
        b.results = SUCCESS
        step = self.create_fake_build_step()
        step.flunkOnWarnings = True
        terminate = b.stepDone(WARNINGS, step)
        self.assertFalse(terminate.result)
        self.assertEqual(b.results, FAILURE)

    def testStepDoneHaltOnFailureFlunkOnWarnings(self):
        b = self.build
        b.results = SUCCESS
        step = self.create_fake_build_step()
        step.flunkOnWarnings = True
        self.haltOnFailure = True
        terminate = b.stepDone(WARNINGS, step)
        self.assertFalse(terminate.result)
        self.assertEqual(b.results, FAILURE)

    def testStepDoneWarningsDontOverrideFailure(self):
        b = self.build
        b.results = FAILURE
        step = self.create_fake_build_step()
        terminate = b.stepDone(WARNINGS, step)
        self.assertFalse(terminate.result)
        self.assertEqual(b.results, FAILURE)

    def testStepDoneRetryOverridesAnythingElse(self):
        b = self.build
        b.results = RETRY
        step = self.create_fake_build_step()
        step.alwaysRun = True
        b.stepDone(WARNINGS, step)
        b.stepDone(FAILURE, step)
        b.stepDone(SUCCESS, step)
        terminate = b.stepDone(EXCEPTION, step)
        self.assertTrue(terminate.result)
        self.assertEqual(b.results, RETRY)

    def test_getSummaryStatistic(self):
        b = self.build

        b.executedSteps = [BuildStep(), BuildStep(), BuildStep()]
        b.executedSteps[0].setStatistic('casualties', 7)
        b.executedSteps[2].setStatistic('casualties', 4)

        add = operator.add
        self.assertEqual(b.getSummaryStatistic('casualties', add), 11)
        self.assertEqual(b.getSummaryStatistic('casualties', add, 10), 21)

    def create_fake_steps(self, names):
        steps = []

        def create_fake_step(name):
            step = self.create_fake_build_step()
            step.name = name
            return step

        for name in names:
            step = create_fake_step(name)
            steps.append(step)
        return steps

    @defer.inlineCallbacks
    def test_start_build_sets_properties(self):
        b = self.build
        b.setProperty("foo", "bar", "test")

        step = create_step_from_step_or_factory(self.create_fake_build_step())
        b.setStepFactories([FakeStepFactory(step)])

        yield b.startBuild(self.workerforbuilder)
        self.assertEqual(b.results, SUCCESS)

        properties = yield self.master.data.get(('builds', 1, 'properties'))
        del properties['builddir']  # contains per-platform data
        self.assertEqual(
            properties,
            {
                'basedir': ('/wrk', 'Worker'),
                'branch': (None, 'Build'),
                'buildnumber': (1, 'Build'),
                'codebase': ('', 'Build'),
                'foo': ('bar', 'test'),  # custom property
                'owners': (['me'], 'Build'),
                'project': ('', 'Build'),
                'repository': ('', 'Build'),
                'revision': ('12345', 'Build'),
            },
        )

    @defer.inlineCallbacks
    def testAddStepsAfterCurrentStep(self):
        b = self.build

        steps = self.create_fake_steps(["a", "b", "c"])

        def startStepB(*args, **kw):
            new_steps = self.create_fake_steps(["d", "e"])
            b.addStepsAfterCurrentStep([FakeStepFactory(s) for s in new_steps])
            return SUCCESS

        steps[1].startStep = startStepB
        b.setStepFactories([FakeStepFactory(s) for s in steps])

        yield b.startBuild(self.workerforbuilder)
        self.assertEqual(b.results, SUCCESS)
        expected_names = ["a", "b", "d", "e", "c"]
        executed_names = [s.name for s in b.executedSteps]
        self.assertEqual(executed_names, expected_names)

    @defer.inlineCallbacks
    def testAddStepsAfterLastStep(self):
        b = self.build

        steps = self.create_fake_steps(["a", "b", "c"])

        def startStepB(*args, **kw):
            new_steps = self.create_fake_steps(["d", "e"])
            b.addStepsAfterLastStep([FakeStepFactory(s) for s in new_steps])
            return SUCCESS

        steps[1].startStep = startStepB
        b.setStepFactories([FakeStepFactory(s) for s in steps])

        yield b.startBuild(self.workerforbuilder)
        self.assertEqual(b.results, SUCCESS)
        expected_names = ["a", "b", "c", "d", "e"]
        executed_names = [s.name for s in b.executedSteps]
        self.assertEqual(executed_names, expected_names)

    def testStepNamesUnique(self):
        # if the step names are unique they should remain unchanged
        b = self.build

        steps = self.create_fake_steps(["clone", "command", "clean"])
        b.setStepFactories([FakeStepFactory(s) for s in steps])

        b.startBuild(self.workerforbuilder)
        self.assertEqual(b.results, SUCCESS)
        expected_names = ["clone", "command", "clean"]
        executed_names = [s.name for s in b.executedSteps]
        self.assertEqual(executed_names, expected_names)

    def testStepNamesDuplicate(self):
        b = self.build

        steps = self.create_fake_steps(["stage", "stage", "stage"])
        b.setStepFactories([FakeStepFactory(s) for s in steps])

        b.startBuild(self.workerforbuilder)
        self.assertEqual(b.results, SUCCESS)
        expected_names = ["stage", "stage_1", "stage_2"]
        executed_names = [s.name for s in b.executedSteps]
        self.assertEqual(executed_names, expected_names)

    def testStepNamesDuplicateAfterAdd(self):
        b = self.build

        steps = self.create_fake_steps(["a", "b", "c"])

        def startStepB(*args, **kw):
            new_steps = self.create_fake_steps(["c", "c"])
            b.addStepsAfterCurrentStep([FakeStepFactory(s) for s in new_steps])
            return SUCCESS

        steps[1].startStep = startStepB
        b.setStepFactories([FakeStepFactory(s) for s in steps])

        b.startBuild(self.workerforbuilder)
        self.assertEqual(b.results, SUCCESS)
        expected_names = ["a", "b", "c", "c_1", "c_2"]
        executed_names = [s.name for s in b.executedSteps]
        self.assertEqual(executed_names, expected_names)

    @defer.inlineCallbacks
    def testGetUrl(self):
        self.build.number = 3
        url = yield self.build.getUrl()
        self.assertEqual(url, 'http://localhost:8080/#/builders/83/builds/3')

    @defer.inlineCallbacks
    def testGetUrlForVirtualBuilder(self):
        # Let's fake a virtual builder
        yield self.master.db.insert_test_data([
            fakedb.Builder(id=108, name='wilma'),
        ])
        self.builder._builders['wilma'] = 108
        self.build.setProperty('virtual_builder_name', 'wilma', 'Build')
        self.build.setProperty('virtual_builder_tags', ['_virtual_'])
        self.build.number = 33
        url = yield self.build.getUrl()
        self.assertEqual(url, 'http://localhost:8080/#/builders/108/builds/33')

    def test_active_builds_metric(self):
        """
        The number of active builds is increased when a build starts
        and decreased when it finishes.
        """
        b = self.build

        controller, step_factory = makeControllableStepFactory()
        b.setStepFactories([step_factory])

        observer = MetricLogObserver()
        observer.enable()
        self.addCleanup(observer.disable)

        def get_active_builds():
            return observer.asDict()['counters'].get('active_builds', 0)

        self.assertEqual(get_active_builds(), 0)

        b.startBuild(self.workerforbuilder)

        self.assertEqual(get_active_builds(), 1)

        controller.finishStep(SUCCESS)

        self.assertEqual(get_active_builds(), 0)

    def test_active_builds_metric_failure(self):
        """
        The number of active builds is increased when a build starts
        and decreased when it finishes..
        """
        b = self.build

        b.setStepFactories([FailingStepFactory()])

        observer = MetricLogObserver()
        observer.enable()
        self.addCleanup(observer.disable)

        def get_active_builds():
            return observer.asDict()['counters'].get('active_builds', 0)

        self.assertEqual(get_active_builds(), 0)

        b.startBuild(self.workerforbuilder)

        self.flushLoggedErrors(TestException)

        self.assertEqual(get_active_builds(), 0)

    def test_build_env_mutation_not_propagated(self):
        builder = FakeBuilder(self.master)
        builder.config.env['PATH'] = ['/a/b/c', '/d/e/f']
        build = Build(self.build.requests, builder)

        # build correctly inherited
        self.assertEqual(build.env['PATH'], ['/a/b/c', '/d/e/f'])
        # should NOT be the same object
        self.assertIsNot(build.env['PATH'], builder.config.env['PATH'])

        # mutate build.env
        build.env['PATH'].insert(0, '/x/y/z')

        self.assertEqual(build.env['PATH'], ['/x/y/z', '/a/b/c', '/d/e/f'])
        self.assertEqual(builder.config.env['PATH'], ['/a/b/c', '/d/e/f'])


class TestMultipleSourceStamps(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self)
        self.builder = FakeBuilder(self.master)

        r = FakeRequest()
        s1 = FakeSource()
        s1.repository = "repoA"
        s1.codebase = "A"
        s1.changes = [FakeChange(10), FakeChange(11)]
        s1.revision = "12345"
        s2 = FakeSource()
        s2.repository = "repoB"
        s2.codebase = "B"
        s2.changes = [FakeChange(12), FakeChange(13)]
        s2.revision = "67890"
        s3 = FakeSource()
        s3.repository = "repoC"
        # no codebase defined
        s3.changes = [FakeChange(14), FakeChange(15)]
        s3.revision = "111213"
        r.sources.extend([s1, s2, s3])

        self.build = Build([r], self.builder)

    def test_buildReturnSourceStamp(self):
        """
        Test that a build returns the correct sourcestamp
        """
        source1 = self.build.getSourceStamp("A")
        source2 = self.build.getSourceStamp("B")

        self.assertEqual([source1.repository, source1.revision], ["repoA", "12345"])
        self.assertEqual([source2.repository, source2.revision], ["repoB", "67890"])

    def test_buildReturnSourceStamp_empty_codebase(self):
        """
        Test that a build returns the correct sourcestamp if codebase is empty
        """
        codebase = ''
        source3 = self.build.getSourceStamp(codebase)
        self.assertTrue(source3 is not None)
        self.assertEqual([source3.repository, source3.revision], ["repoC", "111213"])


class TestBuildBlameList(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self)
        self.builder = FakeBuilder(self.master)

        self.sourceByMe = FakeSource()
        self.sourceByMe.repository = "repoA"
        self.sourceByMe.codebase = "A"
        self.sourceByMe.changes = [FakeChange(10), FakeChange(11)]
        self.sourceByMe.changes[0].who = "me"
        self.sourceByMe.changes[1].who = "me"

        self.sourceByHim = FakeSource()
        self.sourceByHim.repository = "repoB"
        self.sourceByHim.codebase = "B"
        self.sourceByHim.changes = [FakeChange(12), FakeChange(13)]
        self.sourceByHim.changes[0].who = "him"
        self.sourceByHim.changes[1].who = "him"

        self.patchSource = FakeSource()
        self.patchSource.repository = "repoB"
        self.patchSource.codebase = "B"
        self.patchSource.changes = []
        self.patchSource.revision = "67890"
        self.patchSource.patch_info = ("jeff", "jeff's new feature")

    def test_blamelist_for_changes(self):
        r = FakeRequest()
        r.sources.extend([self.sourceByMe, self.sourceByHim])
        build = Build([r], self.builder)
        blamelist = build.blamelist()
        self.assertEqual(blamelist, ['him', 'me'])

    def test_blamelist_for_patch(self):
        r = FakeRequest()
        r.sources.extend([self.patchSource])
        build = Build([r], self.builder)
        blamelist = build.blamelist()
        # If no patch is set, author will not be est
        self.assertEqual(blamelist, [])


class TestSetupProperties_MultipleSources(TestReactorMixin, unittest.TestCase):
    """
    Test that the property values, based on the available requests, are
    initialized properly
    """

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.props = {}
        self.r = FakeRequest()
        self.r.sources = []
        self.r.sources.append(FakeSource())
        self.r.sources[0].changes = [FakeChange()]
        self.r.sources[0].repository = "http://svn-repo-A"
        self.r.sources[0].codebase = "A"
        self.r.sources[0].branch = "develop"
        self.r.sources[0].revision = "12345"
        self.r.sources.append(FakeSource())
        self.r.sources[1].changes = [FakeChange()]
        self.r.sources[1].repository = "http://svn-repo-B"
        self.r.sources[1].codebase = "B"
        self.r.sources[1].revision = "34567"
        self.builder = FakeBuilder((yield fakemaster.make_master(self, wantData=True)))
        self.build = Build([self.r], self.builder)
        self.build.setStepFactories([])
        # record properties that will be set
        self.build.properties.setProperty = self.setProperty

    def setProperty(self, n, v, s, runtime=False):
        if s not in self.props:
            self.props[s] = {}
        if not self.props[s]:
            self.props[s] = {}
        self.props[s][n] = v

    def test_sourcestamp_properties_not_set(self):
        Build.setupBuildProperties(self.build.getProperties(), [self.r], self.r.sources)
        self.assertNotIn("codebase", self.props["Build"])
        self.assertNotIn("revision", self.props["Build"])
        self.assertNotIn("branch", self.props["Build"])
        self.assertNotIn("project", self.props["Build"])
        self.assertNotIn("repository", self.props["Build"])


class TestSetupProperties_SingleSource(TestReactorMixin, unittest.TestCase):
    """
    Test that the property values, based on the available requests, are
    initialized properly
    """

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.props = {}
        self.r = FakeRequest()
        self.r.sources = []
        self.r.sources.append(FakeSource())
        self.r.sources[0].changes = [FakeChange()]
        self.r.sources[0].repository = "http://svn-repo-A"
        self.r.sources[0].codebase = "A"
        self.r.sources[0].branch = "develop"
        self.r.sources[0].revision = "12345"
        self.builder = FakeBuilder((yield fakemaster.make_master(self, wantData=True)))
        self.build = Build([self.r], self.builder)
        self.build.setStepFactories([])
        # record properties that will be set
        self.build.properties.setProperty = self.setProperty

    def setProperty(self, n, v, s, runtime=False):
        if s not in self.props:
            self.props[s] = {}
        if not self.props[s]:
            self.props[s] = {}
        self.props[s][n] = v

    def test_properties_codebase(self):
        Build.setupBuildProperties(self.build.getProperties(), [self.r], self.r.sources)
        codebase = self.props["Build"]["codebase"]
        self.assertEqual(codebase, "A")

    def test_properties_repository(self):
        Build.setupBuildProperties(self.build.getProperties(), [self.r], self.r.sources)
        repository = self.props["Build"]["repository"]
        self.assertEqual(repository, "http://svn-repo-A")

    def test_properties_revision(self):
        Build.setupBuildProperties(self.build.getProperties(), [self.r], self.r.sources)
        revision = self.props["Build"]["revision"]
        self.assertEqual(revision, "12345")

    def test_properties_branch(self):
        Build.setupBuildProperties(self.build.getProperties(), [self.r], self.r.sources)
        branch = self.props["Build"]["branch"]
        self.assertEqual(branch, "develop")

    def test_property_project(self):
        Build.setupBuildProperties(self.build.getProperties(), [self.r], self.r.sources)
        project = self.props["Build"]["project"]
        self.assertEqual(project, '')



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import os
from unittest import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.worker import local


class TestLocalWorker(TestReactorMixin, unittest.TestCase):
    try:
        from buildbot_worker.bot import LocalWorker as _
    except ImportError:
        skip = "buildbot-worker package is not installed"

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantDb=True, wantData=True)
        self.botmaster = self.master.botmaster
        self.workers = self.master.workers

    def createWorker(self, name='bot', attached=False, configured=True, **kwargs):
        worker = local.LocalWorker(name, **kwargs)
        if configured:
            worker.setServiceParent(self.workers)
        return worker

    @defer.inlineCallbacks
    def test_reconfigService_attrs(self):
        old = self.createWorker(
            'bot',
            max_builds=2,
            notify_on_missing=['me@me.com'],
            missing_timeout=120,
            properties={'a': 'b'},
        )
        new = self.createWorker(
            'bot',
            configured=False,
            max_builds=3,
            notify_on_missing=['her@me.com'],
            missing_timeout=121,
            workdir=os.path.abspath('custom'),
            properties={'a': 'c'},
        )

        old.updateWorker = mock.Mock(side_effect=lambda: defer.succeed(None))
        yield old.startService()
        self.assertEqual(old.remote_worker.bot.basedir, os.path.abspath('basedir/workers/bot'))

        yield old.reconfigServiceWithSibling(new)

        self.assertEqual(old.max_builds, 3)
        self.assertEqual(old.notify_on_missing, ['her@me.com'])
        self.assertEqual(old.missing_timeout, 121)
        self.assertEqual(old.properties.getProperty('a'), 'c')
        self.assertEqual(old.registration.updates, ['bot'])
        self.assertTrue(old.updateWorker.called)
        # make sure that we can provide an absolute path
        self.assertEqual(old.remote_worker.bot.basedir, os.path.abspath('custom'))
        yield old.stopService()

    @defer.inlineCallbacks
    def test_workerinfo(self):
        wrk = self.createWorker(
            'bot',
            max_builds=2,
            notify_on_missing=['me@me.com'],
            missing_timeout=120,
            properties={'a': 'b'},
        )
        yield wrk.startService()
        info = yield wrk.conn.remoteGetWorkerInfo()
        self.assertIn("worker_commands", info)
        yield wrk.stopService()
